# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from torch.autograd import Function

from . import DOMAIN_STRING


class QuantHardTanhFn(Function):

    @staticmethod
    def symbolic(g, input, qnt_type, thres, bias, scale):
        if qnt_type == "BIPOLAR":
            return g.op(
                f'{DOMAIN_STRING}::MultiThreshold',
                input,
                thres,
                out_dtype_s=qnt_type,
                out_scale_f=2.0,
                out_bias_f=-1.0)
        else:
            ret = g.op(f'{DOMAIN_STRING}::MultiThreshold', input, thres, out_dtype_s=qnt_type)
            if bias is not None:
                ret = g.op('Add', ret, bias)
            if scale is not None:
                ret = g.op('Mul', ret, scale)
            return ret

    @staticmethod
    def forward(ctx, input, qnt_type, thres, bias, scale):
        return input.clamp(0)


class QuantReLUFn(Function):

    @staticmethod
    def symbolic(g, input, qnt_type, thres, bias, scale):
        ret = g.op(f'{DOMAIN_STRING}::MultiThreshold', input, thres, out_dtype_s=qnt_type)
        if scale is not None:
            ret = g.op('Mul', ret, scale)
        return ret

    @staticmethod
    def forward(ctx, input, qnt_type, thres, bias, scale):
        return input.clamp(0)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from functools import reduce
from operator import mul
from typing import Optional, Tuple, Type, Union

import torch
from torch import Tensor
from torch.nn import AdaptiveAvgPool2d
from torch.nn import AvgPool2d

from brevitas.function.ops import max_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.inject.defaults import RoundTo8bit
from brevitas.quant_tensor import QuantTensor

from .mixin.acc import AccQuantType
from .mixin.acc import TruncMixin
from .mixin.base import QuantLayerMixin


class TruncAvgPool2d(TruncMixin, QuantLayerMixin, AvgPool2d):
    """
    Quantized AvgPool2d variant that replaces the division step in the average with a right shift
    to the target bit-width through a rounding or truncation quantizer.
    Requires a QuantTensor as input and preserves the scale of the input in the output.
    """

    def __init__(
            self,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = None,
            trunc_quant: Optional[AccQuantType] = RoundTo8bit,
            return_quant_tensor: bool = True,
            **kwargs):
        AvgPool2d.__init__(self, kernel_size=kernel_size, stride=stride)
        QuantLayerMixin.__init__(self, return_quant_tensor)
        TruncMixin.__init__(self, trunc_quant=trunc_quant, **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return True

    @property
    def _avg_scaling(self):
        if isinstance(self.kernel_size, tuple):
            return self.kernel_size[0] * self.kernel_size[1]
        else:
            return self.kernel_size * self.kernel_size

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            return self.export_handler(x.value)
        x = x.set(value=super(TruncAvgPool2d, self).forward(x.value))
        if self.is_trunc_quant_enabled:
            assert x.is_not_none  # check input quant tensor is filled with values
            # remove avg scaling
            rescaled_value = x.value * self._avg_scaling
            x = x.set(value=rescaled_value)
            x = x.set(bit_width=self.max_acc_bit_width(x.bit_width))
            x = self.trunc_quant(x)
        return self.pack_output(x)

    def max_acc_bit_width(self, input_bit_width):
        max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
        max_uint_output = max_uint_input * self._avg_scaling
        max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
        return max_output_bit_width


class TruncAdaptiveAvgPool2d(TruncMixin, QuantLayerMixin, AdaptiveAvgPool2d):
    """
    Quantized AdaptiveAvgPool2d variant that replaces the division step in the average with a right shift
    to the target bit-width through a truncation quantizer.
    Requires a QuantTensor as input and preserves the scale of the input in the output.
    """

    def __init__(
            self,
            output_size: Union[int, Tuple[int, int]],
            trunc_quant: Optional[AccQuantType] = RoundTo8bit,
            return_quant_tensor: bool = True,
            cache_kernel_size_stride: bool = True,
            **kwargs):
        AdaptiveAvgPool2d.__init__(self, output_size=output_size)
        QuantLayerMixin.__init__(self, return_quant_tensor)
        TruncMixin.__init__(self, trunc_quant=trunc_quant, **kwargs)
        self.cache_kernel_size_stride = cache_kernel_size_stride
        self._cached_kernel_size = None
        self._cached_kernel_stride = None

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return True

    @property
    def padding(self):
        return 0

    @property
    def kernel_size(self):
        return self._cached_kernel_size

    @property
    def stride(self):
        return self._cached_kernel_stride

    def compute_kernel_size_stride(self, input_shape, output_shape):
        kernel_size_list = []
        stride_list = []
        for inp, out in zip(input_shape, output_shape):
            stride = inp // out
            kernel_size = inp - (out - 1) * stride
            kernel_size_list.append(kernel_size)
            stride_list.append(stride)
        return kernel_size_list, stride_list

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        # shortcut execution through the export impl during export
        if self.export_mode:
            out = self.export_handler(x.value)
            self._set_global_is_quant_layer(False)
            return out
        y = x.set(value=super(TruncAdaptiveAvgPool2d, self).forward(x.value))
        k_size, stride = self.compute_kernel_size_stride(x.value.shape[2:], y.value.shape[2:])
        if self.cache_kernel_size_stride:
            self._cached_kernel_size = k_size
            self._cached_kernel_stride = stride
        if self.is_trunc_quant_enabled:
            assert y.is_not_none  # check input quant tensor is filled with values
            reduce_size = reduce(mul, k_size, 1)
            rescaled_value = y.value * reduce_size  # remove avg scaling
            y = y.set(value=rescaled_value)
            y = y.set(bit_width=self.max_acc_bit_width(y.bit_width, reduce_size))
            y = self.trunc_quant(y)
        return self.pack_output(y)

    def max_acc_bit_width(self, input_bit_width, reduce_size):
        max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
        max_uint_output = max_uint_input * reduce_size
        max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
        return max_output_bit_width



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from copy import copy

import torch
from torch import Tensor

from brevitas.export.common.handler.base import QuantAxisMixin
from brevitas.export.common.handler.qcdq import DQMixin
from brevitas.export.common.handler.qcdq import QCDQActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQBiasQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQDecoupledWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQMixin
from brevitas.export.common.handler.qcdq import QCDQTruncQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import ZeroPointHandlerMixin
from brevitas.export.onnx.handler import ONNXBaseHandler
from brevitas.export.onnx.handler import QuantLSTMLayerHandler

from ..function import DequantizeLinearFn
from ..function import IntClipFn
from ..function import QuantizeLinearFn


class StdDQONNXMixin(DQMixin, ABC):

    def dequantize_fn(self, x, scale, zero_point, axis):
        return DequantizeLinearFn.apply(x, scale, zero_point, axis)

    @property
    def flatten_dequantize_params(self):
        return True

    @property
    def itemize_quantize_scalar_params(self):
        return False


class StdQCDQONNXMixin(QCDQMixin, StdDQONNXMixin, ABC):

    @property
    def clip_over_integers(self):
        return True

    @classmethod
    def int8_dtype(cls):
        return torch.int8

    @classmethod
    def uint8_dtype(cls):
        return torch.uint8

    @classmethod
    def int32_dtype(cls):
        return torch.int32

    def validate(self, module):
        self.validate_8b_bit_width(module.bit_width(), le_then=True)
        assert module.bit_width() > 1., 'Binary quant not supported'
        assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        return QuantizeLinearFn.apply(x, scale, zero_point, dtype, axis)

    def clip_fn(self, x, min_val, max_val):
        return IntClipFn.apply(x, min_val, max_val)


class StdQCDQONNXWeightQuantProxyHandler(StdQCDQONNXMixin,
                                         QCDQWeightQuantProxyHandlerMixin,
                                         ONNXBaseHandler):
    pass


class StdQCDQONNXDecoupledWeightQuantProxyHandler(StdQCDQONNXMixin,
                                                  QCDQDecoupledWeightQuantProxyHandlerMixin,
                                                  ONNXBaseHandler):
    pass


class StdQCDQONNXActQuantProxyHandler(StdQCDQONNXMixin,
                                      QCDQActQuantProxyHandlerMixin,
                                      ONNXBaseHandler):
    pass


class StdQCDQONNXBiasQuantProxyHandler(StdDQONNXMixin,
                                       QCDQBiasQuantProxyHandlerMixin,
                                       ONNXBaseHandler):
    pass


class StdQCDQONNXTruncQuantProxyHandler(StdQCDQONNXMixin,
                                        QCDQTruncQuantProxyHandlerMixin,
                                        ONNXBaseHandler):
    pass


class StdQCDQONNXQuantLSTMLayerHandler(QuantLSTMLayerHandler):

    def quantized_cell_symbolic_execution(
            self,
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output):
        raise RuntimeError(
            "Quantized LSTM cell is not supported for ONNX QCDQ "
            "(weights only quantization is). Use export_qonnx.")



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .hadamard_classifier import HadamardClassifier
from .quant_accumulator import ClampQuantAccumulator
from .quant_accumulator import TruncQuantAccumulator
from .quant_activation import QuantHardTanh
from .quant_activation import QuantIdentity
from .quant_activation import QuantReLU
from .quant_activation import QuantSigmoid
from .quant_activation import QuantTanh
from .quant_avg_pool import TruncAdaptiveAvgPool2d
from .quant_avg_pool import TruncAvgPool2d
from .quant_bn import BatchNorm1dToQuantScaleBias
from .quant_bn import BatchNorm2dToQuantScaleBias
from .quant_conv import QuantConv1d
from .quant_conv import QuantConv2d
from .quant_convtranspose import QuantConvTranspose1d
from .quant_convtranspose import QuantConvTranspose2d
from .quant_dropout import QuantDropout
from .quant_eltwise import QuantCat
from .quant_eltwise import QuantEltwiseAdd
from .quant_embedding import QuantEmbedding
from .quant_linear import QuantLinear
from .quant_max_pool import QuantMaxPool1d
from .quant_max_pool import QuantMaxPool2d
from .quant_mha import QuantMultiheadAttention
from .quant_rnn import QuantLSTM
from .quant_rnn import QuantRNN
from .quant_scale_bias import QuantScaleBias
from .quant_scale_bias import ScaleBias
from .quant_upsample import QuantUpsample
from .quant_upsample import QuantUpsamplingBilinear2d
from .quant_upsample import QuantUpsamplingNearest2d
from .target import flexml



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.function_wrapper import TensorClamp
from brevitas.quant.base import *
from brevitas.quant.solver.act import ActQuantSolver
from brevitas.quant.solver.bias import BiasQuantSolver
from brevitas.quant.solver.trunc import TruncQuantSolver
from brevitas.quant.solver.weight import WeightQuantSolver

__all__ = [
    'IntBias',
    'Int8Bias',
    'Int16Bias',
    'Int24Bias',
    'Int32Bias',
    'Int8BiasPerTensorFloatInternalScaling',
    'Int8ActPerTensorFloatMinMaxInit',
    'Uint8ActPerTensorFloatMaxInit',
    'Int8ActPerTensorFloat',
    'Int8WeightPerTensorFloat',
    'Int8WeightPerChannelFloat',
    'Uint8ActPerTensorFloat',
    'TruncTo8bit',
    'RoundTo8bit',
    'Int4WeightPerTensorFloatDecoupled',
    'Int8WeightPerChannelFloatDecoupled',
    'Uint8ActPerTensorFloatBatchQuant1d',
    'Int8ActPerTensorFloatBatchQuant1d',
    'Uint8ActPerTensorFloatBatchQuant2d',
    'Int8ActPerTensorFloatBatchQuant2d']


class Int8ActPerTensorFloatMinMaxInit(IntQuant,
                                      ParamMinMaxInitScaling,
                                      PerTensorFloatScaling8bit,
                                      ActQuantSolver):
    """
    8-bit per-tensor signed int activations quantizer with learned floating-point scale factor
    initialized from user-defined min and max values.

    Examples:
        >>> from brevitas.nn import QuantHardTanh
        >>> act = QuantHardTanh(act_quant=Int8ActPerTensorFloatMinMaxInit, min_val=-.5, max_val=.5)
        >>> act.quant_act_scale() * - 128
        tensor(-0.5000, grad_fn=<MulBackward0>)
        >>> act.quant_act_scale() * 127
        tensor(0.4961, grad_fn=<MulBackward0>)
    """
    pass


class Uint8ActPerTensorFloatMaxInit(UintQuant,
                                    ParamMinMaxInitScaling,
                                    PerTensorFloatScaling8bit,
                                    ActQuantSolver):
    """
    8-bit per-tensor unsigned int activations quantizer with learned floating-point scale factor
    initialized from a user-defined max val.

    Examples:
        >>> from brevitas.nn import QuantHardTanh
        >>> act = QuantHardTanh(act_quant=Uint8ActPerTensorFloatMaxInit, max_val=.5)
        >>> act.quant_act_scale() * 255
        tensor(0.5000, grad_fn=<MulBackward0>)
    """
    min_val = 0.0


class IntBias(IntQuant, BiasQuantSolver):
    """
    Signed int bias quantizer with bit-width and scale factor equal to the bit-width and the scale
    factor of the accumulator the bias is added to.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=IntBias)
    """
    tensor_clamp_impl = TensorClamp
    requires_input_scale = True
    requires_input_bit_width = True


class Int8Bias(IntBias):
    """
    8-bit signed int bias quantizer with scale factor equal to the scale factor of the accumulator
    the bias is added to, so typically quant_input_scale * quant_weight_scale.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=Int8Bias)
    """
    bit_width = 8
    requires_input_bit_width = False


class Int16Bias(IntBias):
    """
    16-bit signed int bias quantizer with scale factor equal to the scale factor of the accumulator
    the bias is added to, so typically quant_input_scale * quant_weight_scale.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=Int16Bias)
    """
    bit_width = 16
    requires_input_bit_width = False


class Int24Bias(IntBias):
    """
    24-bit signed int bias quantizer with scale factor equal to the scale factor of the accumulator
    the bias is added to, so typically quant_input_scale * quant_weight_scale.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=Int16Bias)
    """
    bit_width = 24
    requires_input_bit_width = False


class Int32Bias(IntBias):
    """
    32-bit signed int bias quantizer with scale factor equal to the scale factor of the accumulator
    the bias is added to, so typically quant_input_scale * quant_weight_scale.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=Int16Bias)
    """
    bit_width = 32
    requires_input_bit_width = False


class Int8BiasPerTensorFloatInternalScaling(IntQuant,
                                            MaxStatsScaling,
                                            PerTensorFloatScaling8bit,
                                            BiasQuantSolver):
    """
    8-bit per-tensor signed int bias quantizer with floating-point scale factor computed from
    backpropagated statistics of the bias tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=True, bias_quant=Int8BiasPerTensorFloatInternalScaling)
    """
    requires_input_scale = False
    requires_input_bit_width = False


class Int8WeightPerTensorFloat(NarrowIntQuant,
                               MaxStatsScaling,
                               PerTensorFloatScaling8bit,
                               WeightQuantSolver):
    """
    8-bit narrow per-tensor signed int weight quantizer with per-tensor floating-point scale factor computed
    from backpropagated statistics of the weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=Int8WeightPerTensorFloat)
    """
    pass


class Int8WeightPerChannelFloat(NarrowIntQuant,
                                MaxStatsScaling,
                                PerChannelFloatScaling8bit,
                                WeightQuantSolver):
    """
    8-bit narrow per-tensor signed int weight quantizer with per-channel floating-point scale factor computed
    from backpropagated statistics of the weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=Int8WeightPerChannelFloat)
    """
    pass


class Int8ActPerTensorFloat(IntQuant,
                            ParamFromRuntimePercentileScaling,
                            PerTensorFloatScaling8bit,
                            ActQuantSolver):
    """
    8-bit per-tensor signed int activations quantizer with learned floating-point scale factor
    initialized from runtime statistics.

    Examples:
        >>> from brevitas.nn import QuantIdentity
        >>> act = QuantIdentity(act_quant=Int8ActPerTensorFloat)
    """
    pass


class Uint8ActPerTensorFloat(UintQuant,
                             ParamFromRuntimePercentileScaling,
                             PerTensorFloatScaling8bit,
                             ActQuantSolver):
    """
    8-bit per-tensor unsigned int activations quantizer with learned floating-point scale factor
    initialized from runtime statistics.

    Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=Uint8ActPerTensorFloat)
    """
    pass


class TruncTo8bit(TruncQuantSolver):
    """
    8-bit signed int truncator that preserves the input scale factor and zero-point.

    Examples:
        >>> from brevitas.nn import TruncAvgPool2d
        >>> pool = TruncAvgPool2d(kernel_size=(3, 3), trunc_quant=TruncTo8bit)
    """
    bit_width = 8
    quant_type = 'int'
    bit_width_impl_type = 'const'
    float_to_int_impl_type = 'floor'


class RoundTo8bit(TruncQuantSolver):
    """
    8-bit signed int truncator with rounding that preserves the input scale factor and zero-point.

    Examples:
        >>> from brevitas.nn import TruncAvgPool2d
        >>> pool = TruncAvgPool2d(kernel_size=(3, 3), trunc_quant=RoundTo8bit)
    """
    bit_width = 8
    quant_type = 'int'
    bit_width_impl_type = 'const'
    float_to_int_impl_type = 'round'


class Int4WeightPerTensorFloatDecoupled(WeightPerTensorFloatDecoupledL2Param):
    """
    Experimental narrow per-tensor signed int weight quantizer with decoupled L2,inf
    normalization and learned scaling. Especially suited for the challenging scenario of
    per-tensor low bit-width quantization of depthwise separable weights when retraining from a
    pretrained floating-point model.

    Examples:
        >>> from brevitas.nn import QuantConv2d
        >>> m = QuantConv2d(4, 4, 3, groups=4, weight_quant=Int4WeightPerTensorFloatDecoupled)
    """
    bit_width = 4


class Int8WeightPerChannelFloatDecoupled(WeightPerChannelFloatDecoupled):
    """
    Experimental narrow per-channel signed int weight quantizer with decoupled Linf
    normalization and learned scaling.

    Examples:
        >>> from brevitas.nn import QuantConv2d
        >>> m = QuantConv2d(4, 4, 3, weight_quant=Int8WeightPerChannelFloatDecoupled)
    """
    bit_width = 8


class Uint8ActPerTensorFloatBatchQuant2d(UintQuant,
                                         BatchQuantStatsScaling2d,
                                         PerTensorFloatScaling8bit,
                                         ActQuantSolver):
    """
    8-bit symmetric per-tensor signed int activations quantizer with affine scale based on s * mean(max_i(abs(x))), s learned,
    with i the output channel dimension of a 4-d tensor with shape (N, C, H, W).
    Statistics are accumulated with a true average rather than a moving exponential one.
    Statistics should be recomputed over calibration data at the end of training, as done in:

    https://proceedings.neurips.cc/paper/2021/hash/08aee6276db142f4b8ac98fb8ee0ed1b-Abstract.html
    Bai, Haoping, et al. "Batchquant: Quantized-for-all architecture search with robust quantizer."
    Advances in Neural Information Processing Systems 34 (2021): 1074-1085.

    Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=Uint8ActPerTensorFloatBatchQuant2d)
    """
    pass


class Int8ActPerTensorFloatBatchQuant2d(IntQuant,
                                        BatchQuantStatsScaling2d,
                                        PerTensorFloatScaling8bit,
                                        ActQuantSolver):
    """
    8-bit symmetric per-tensor unsigned int activations quantizer with affine scale based on s * mean(max_i(abs(x))), s learned,
    with i the output channel dimension of a 4-d tensor with shape (N, C, H, W).
    Statistics are accumulated with a true average rather than a moving exponential one.
    Statistics should be recomputed over calibration data at the end of training, as done in:

    https://proceedings.neurips.cc/paper/2021/hash/08aee6276db142f4b8ac98fb8ee0ed1b-Abstract.html
    Bai, Haoping, et al. "Batchquant: Quantized-for-all architecture search with robust quantizer."
    Advances in Neural Information Processing Systems 34 (2021): 1074-1085.

    Examples:
        >>> from brevitas.nn import QuantIdentity
        >>> act = QuantIdentity(act_quant=Int8ActPerTensorFloatBatchQuant2d)
    """
    pass


class Uint8ActPerTensorFloatBatchQuant1d(UintQuant,
                                         BatchQuantStatsScaling1d,
                                         PerTensorFloatScaling8bit,
                                         ActQuantSolver):
    """
    8-bit symmetric per-tensor signed int activations quantizer with affine scale based on s * mean(max_i(abs(x))), s learned,
    with i the output channel dimension of a 3-d tensor with shape (N, C, Feat).
    Statistics are accumulated with a true average rather than a moving exponential one.
    Statistics should be recomputed over calibration data at the end of training, as done in:

    https://proceedings.neurips.cc/paper/2021/hash/08aee6276db142f4b8ac98fb8ee0ed1b-Abstract.html
    Bai, Haoping, et al. "Batchquant: Quantized-for-all architecture search with robust quantizer."
    Advances in Neural Information Processing Systems 34 (2021): 1074-1085.

    Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=Uint8ActPerTensorFloatBatchQuant1d)
    """
    pass


class Int8ActPerTensorFloatBatchQuant1d(IntQuant,
                                        BatchQuantStatsScaling1d,
                                        PerTensorFloatScaling8bit,
                                        ActQuantSolver):
    """
    8-bit symmetric per-tensor unsigned int activations quantizer with affine scale based on s * mean(max_i(abs(x))), s learned,
    with i the output channel dimension of a 3-d tensor with shape (N, C, Feat).
    Statistics are accumulated with a true average rather than a moving exponential one.
    Statistics should be recomputed over calibration data at the end of training, as done in:

    https://proceedings.neurips.cc/paper/2021/hash/08aee6276db142f4b8ac98fb8ee0ed1b-Abstract.html
    Bai, Haoping, et al. "Batchquant: Quantized-for-all architecture search with robust quantizer."
    Advances in Neural Information Processing Systems 34 (2021): 1074-1085.

    Examples:
        >>> from brevitas.nn import QuantIdentity
        >>> act = QuantIdentity(act_quant=Int8ActPerTensorFloatBatchQuant1d)
    """
    pass



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Tuple, Union

import torch
from torch import Tensor
from torch.nn import Module

from brevitas.export.manager import _set_layer_export_handler
from brevitas.export.manager import _set_layer_export_mode
from brevitas.export.manager import BaseManager
from brevitas.export.manager import ExportContext
from brevitas.quant_tensor import QuantTensor

from .handler import qF
from .handler.act import PytorchQuantHardTanhHandler
from .handler.act import PytorchQuantIdentityHandler
from .handler.act import PytorchQuantReLUHandler
from .handler.parameter import PytorchQuantConv1dHandler
from .handler.parameter import PytorchQuantConv2dHandler
from .handler.parameter import PytorchQuantLinearHandler
from .handler.pool import PytorchQuantMaxPool1d
from .handler.pool import PytorchQuantMaxPool2d


class TorchQOpManager(BaseManager):
    target_name = 'torch'

    handlers = [
        PytorchQuantMaxPool1d,
        PytorchQuantMaxPool2d,
        PytorchQuantHardTanhHandler,
        PytorchQuantIdentityHandler,
        PytorchQuantReLUHandler,
        PytorchQuantConv1dHandler,
        PytorchQuantConv2dHandler,
        PytorchQuantLinearHandler]

    @classmethod
    def set_export_mode(cls, module: Module, enabled: bool):
        _set_layer_export_mode(module, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_layer_export_handler(cls, module)

    @classmethod
    def export(
            cls,
            module: Module,
            input_shape: Optional[Tuple[int, ...]] = None,
            export_path: Optional[str] = None,
            input_t: Optional[Union[Tensor, QuantTensor]] = None):
        if qF is None:
            raise RuntimeError("torch.nn.quantized.functional cannot be imported.")
        if input_shape is None and input_t is None:
            raise RuntimeError("Export requires to pass in either input_shape or input_t")
        if input_shape is not None and input_t is not None:
            raise RuntimeError("Export accepts either an input shape or an input tensor, not both")
        if input_t is None and input_shape is not None:
            input_t = torch.empty(*input_shape)
        with ExportContext(cls):
            traced_module = cls.jit_inference_trace(module, input_t)
        if export_path is not None:
            traced_module.save(export_path)
        return traced_module



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from torch.autograd import Function

from brevitas.export.onnx import onnx_export_opset

AXIS_OPSET = 13


class DequantizeLinearFn(Function):

    @staticmethod
    def symbolic(g, x, input_scale, input_zero_point, input_axis):
        opset_version = onnx_export_opset()

        if input_axis is not None and opset_version < AXIS_OPSET:
            raise RuntimeError('ONNX Opset 13 is required for per-channel quantization')
        elif input_axis is not None and opset_version >= AXIS_OPSET:
            ret = g.op('DequantizeLinear', x, input_scale, input_zero_point, axis_i=input_axis)
        else:
            ret = g.op('DequantizeLinear', x, input_scale, input_zero_point)
        return ret

    @staticmethod
    def forward(ctx, int_x, input_scale, input_zero_point, input_axis):
        return int_x.float()


class IntClipFn(Function):

    @staticmethod
    def symbolic(g, int_x, min_int_val, max_int_val):
        ret = g.op('Clip', int_x, min_int_val, max_int_val)
        return ret

    @staticmethod
    def forward(ctx, int_x, min_int_val, max_int_val):
        return int_x


class QuantizeLinearFn(Function):

    @staticmethod
    def symbolic(g, x, output_scale, ouput_zero_point, output_dtype, output_axis):
        opset_version = onnx_export_opset()

        if output_axis is not None and opset_version < AXIS_OPSET:
            raise RuntimeError('ONNX Opset 13 is required for per-channel quantization')
        elif output_axis is not None and opset_version >= AXIS_OPSET:
            ret = g.op('QuantizeLinear', x, output_scale, ouput_zero_point, axis_i=output_axis)
        else:
            ret = g.op('QuantizeLinear', x, output_scale, ouput_zero_point)
        return ret

    @staticmethod
    def forward(ctx, x, output_scale, ouput_zero_point, output_dtype, output_axis):
        return x.type(output_dtype)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from torch.nn import Module

from brevitas.export.manager import _set_proxy_export_handler
from brevitas.export.manager import _set_proxy_export_mode
from brevitas.export.manager import _set_recurrent_layer_export_handler
from brevitas.export.manager import _set_recurrent_layer_export_mode
from brevitas.export.onnx.debug import DebugMarkerFunction
from brevitas.export.onnx.function import LSTMCellFn

from ..function import DequantizeLinearFn
from ..function import IntClipFn
from ..function import QuantizeLinearFn
from ..manager import StdONNXBaseManager
from .handler import StdQCDQONNXActQuantProxyHandler
from .handler import StdQCDQONNXBiasQuantProxyHandler
from .handler import StdQCDQONNXDecoupledWeightQuantProxyHandler
from .handler import StdQCDQONNXQuantLSTMLayerHandler
from .handler import StdQCDQONNXTruncQuantProxyHandler
from .handler import StdQCDQONNXWeightQuantProxyHandler


class StdQCDQONNXManager(StdONNXBaseManager):
    target_name = 'StdQCDQONNX'
    dequantize_tracing_input = False

    onnx_passes = [
        # use initializers instead of Constant nodes for fixed params
        "extract_constant_to_initializer",  # remove unused graph inputs & initializers
        "eliminate_unused_initializer"]

    handlers = [
        StdQCDQONNXWeightQuantProxyHandler,
        StdQCDQONNXBiasQuantProxyHandler,
        StdQCDQONNXActQuantProxyHandler,
        StdQCDQONNXDecoupledWeightQuantProxyHandler,
        StdQCDQONNXTruncQuantProxyHandler,
        StdQCDQONNXQuantLSTMLayerHandler]

    custom_fns = [
        DebugMarkerFunction,
        QuantizeLinearFn,
        DequantizeLinearFn,
        IntClipFn,
        LSTMCellFn,]

    @classmethod
    def set_export_mode(cls, model: Module, enabled: bool):
        _set_proxy_export_mode(model, enabled)
        _set_recurrent_layer_export_mode(model, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_proxy_export_handler(cls, module)
        _set_recurrent_layer_export_handler(cls, module)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn import Parameter

import brevitas
from brevitas import config
from brevitas.core.stats import _ParameterListStats
from brevitas.core.stats import DEFAULT_MOMENTUM
from brevitas.core.stats import SCALAR_SHAPE
from brevitas.function import abs_binary_sign_grad

from .utils import inplace_momentum_update
from .utils import inplace_tensor_add
from .utils import StatelessBuffer

__all__ = [
    'ZeroZeroPoint',
    'StatsFromParameterZeroPoint',
    'ParameterFromRuntimeZeroPoint',
    'ParameterZeroPoint']


class ZeroZeroPoint(brevitas.jit.ScriptModule):

    def __init__(self) -> None:
        super(ZeroZeroPoint, self).__init__()
        self.zero_point = StatelessBuffer(torch.tensor(0.0))

    @brevitas.jit.script_method
    def forward(self, x: Tensor, scale: Tensor, bit_width: Tensor) -> Tensor:
        return self.zero_point()


class _ScaleShiftZeroPoint(brevitas.jit.ScriptModule):
    __constants__ = ['quantize_zero_point']

    def __init__(self, int_quant: Module, quantize_zero_point: bool) -> None:
        super(_ScaleShiftZeroPoint, self).__init__()
        self.int_quant = int_quant
        self.quantize_zero_point = quantize_zero_point

    @brevitas.jit.script_method
    def forward(self, zero_point: Tensor, scale: Tensor, bit_width: Tensor) -> Tensor:
        min_int = self.int_quant.min_int(bit_width)
        if self.quantize_zero_point:
            out = self.int_quant.to_int(scale, min_int, bit_width, zero_point)
        else:
            out = zero_point / scale + min_int
        return out


class StatsFromParameterZeroPoint(brevitas.jit.ScriptModule):

    def __init__(
            self,
            int_quant: Module,
            quantize_zero_point: bool,
            zero_point_stats_input_view_shape_impl: Module,
            zero_point_stats_input_concat_dim: int,
            zero_point_stats_impl: Module,
            zero_point_shape: Tuple[int, ...],
            tracked_parameter_list: List[torch.nn.Parameter]) -> None:
        super(StatsFromParameterZeroPoint, self).__init__()
        self.parameter_list_stats = _ParameterListStats(
            zero_point_stats_impl,
            zero_point_shape,
            zero_point_stats_input_view_shape_impl,
            zero_point_stats_input_concat_dim,
            tracked_parameter_list)
        self.scale_shift_zero_point = _ScaleShiftZeroPoint(int_quant, quantize_zero_point)

    @brevitas.jit.script_method
    def forward(self, x: Tensor, scale: Tensor, bit_width: Tensor) -> torch.Tensor:
        stats = self.parameter_list_stats()
        return self.scale_shift_zero_point(-stats, scale, bit_width)


class ParameterFromRuntimeZeroPoint(brevitas.jit.ScriptModule):
    __constants__ = ['stats_permute_dims', 'collect_stats_steps', 'zero_point_shape', 'momentum']

    def __init__(
            self,
            collect_stats_steps: int,
            int_quant: Module,
            quantize_zero_point: bool,
            zero_point_stats_impl: Optional[int],
            zero_point_shape: Tuple[int, ...],
            zero_point_stats_input_view_shape_impl: Module,
            zero_point_stats_momentum: Optional[float] = DEFAULT_MOMENTUM) -> None:
        super(ParameterFromRuntimeZeroPoint, self).__init__()
        assert collect_stats_steps > 0, 'Steps should be more than 0'
        self.collect_stats_steps = collect_stats_steps
        self.counter: int = brevitas.jit.Attribute(0, int)
        self.zero_point_shape = zero_point_shape
        self.stats_input_view_shape_impl = zero_point_stats_input_view_shape_impl
        self.momentum = zero_point_stats_momentum
        self.value = Parameter(torch.full(zero_point_shape, 0.0))
        self.register_buffer('buffer', torch.full(zero_point_shape, 0.0))
        self.zero_point_stats_impl = zero_point_stats_impl
        self.scale_shift_zero_point = _ScaleShiftZeroPoint(int_quant, quantize_zero_point)

    @brevitas.jit.script_method
    def training_forward(self, x) -> Tensor:
        if self.counter < self.collect_stats_steps:
            stats_input = self.stats_input_view_shape_impl(x)
            stats = self.zero_point_stats_impl(stats_input)
            stats = stats.view(self.zero_point_shape)
            new_counter = self.counter + 1
            if self.counter == 0:
                inplace_tensor_add(self.buffer, stats.detach())
            else:
                inplace_momentum_update(
                    self.buffer, stats.detach(), self.momentum, self.counter, new_counter)
            self.counter = new_counter
            # work around find_unusued_parameters=True in DDP
            out = stats + 0. * self.value
        elif self.counter == self.collect_stats_steps:
            inplace_tensor_add(self.value.detach(), self.buffer)
            self.counter = self.counter + 1
            out = self.value
        else:
            out = self.value
        return out

    @brevitas.jit.script_method
    def forward(self, x: Tensor, scale: Tensor, bit_width: Tensor) -> Tensor:
        if self.training:
            out = self.training_forward(x)
        else:
            if self.counter <= self.collect_stats_steps:
                out = self.buffer
            else:
                out = self.value
        out = abs_binary_sign_grad(out)
        out = self.scale_shift_zero_point(out, scale, bit_width)
        return out

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        output_dict = super(ParameterFromRuntimeZeroPoint,
                            self).state_dict(destination, prefix, keep_vars)
        # Avoid saving the buffer
        del output_dict[prefix + 'buffer']
        # Avoid saving the init value
        if self.counter == 0:
            del output_dict[prefix + 'value']
        # Save buffer into value for any non-zero number of collection steps
        elif self.counter <= self.collect_stats_steps:
            output_dict[prefix + 'value'] = self.buffer
        return output_dict

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(ParameterFromRuntimeZeroPoint, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        value_key = prefix + 'value'
        buffer_key = prefix + 'buffer'
        # Buffer is supposed to be always missing
        missing_keys.remove(buffer_key)
        # Pytorch stores training flag as a buffer with JIT enabled
        training_key = prefix + 'training'
        if training_key in missing_keys:
            missing_keys.remove(training_key)
        # disable stats collection when a pretrained value is loaded
        if value_key not in missing_keys:
            self.counter = self.collect_stats_steps + 1
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)


class ParameterZeroPoint(brevitas.jit.ScriptModule):
    __constants__ = ['stats_permute_dims', 'collect_stats_steps', 'momentum']

    def __init__(
            self,
            zero_point_init: Union[float, torch.Tensor],
            int_quant: Module,
            quantize_zero_point: bool,
            zero_point_shape: Tuple[int, ...] = None) -> None:
        super(ParameterZeroPoint, self).__init__()
        if (isinstance(zero_point_init, Tensor) and zero_point_shape is not None and
                zero_point_init.shape != SCALAR_SHAPE and
                zero_point_init.shape != zero_point_shape):
            raise RuntimeError("zero_point_init.shape is non-scalar and != from zero_point_shape.")

        if isinstance(zero_point_init, Tensor):
            zero_point_init = zero_point_init.detach()
        else:
            zero_point_init = torch.tensor(zero_point_init)
        if zero_point_init.shape == SCALAR_SHAPE and zero_point_shape is not None:
            zero_point_init = torch.full(zero_point_shape, zero_point_init)
        self.value = Parameter(zero_point_init)
        self.scale_shift_zero_point = _ScaleShiftZeroPoint(int_quant, quantize_zero_point)

    @brevitas.jit.script_method
    def forward(self, x: Tensor, scale: Tensor, bit_width: Tensor) -> Tensor:
        out = abs_binary_sign_grad(self.value)
        out = self.scale_shift_zero_point(out, scale, bit_width)
        return out

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(ParameterZeroPoint, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        value_key = prefix + 'value'
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from distutils.util import strtobool
import os

try:
    from torch.jit import _enabled
except ImportError:
    from torch.jit._state import _enabled


def env_to_bool(name, default):
    return bool(strtobool(os.environ.get(name, "{}".format(default))))


REINIT_ON_STATE_DICT_LOAD = env_to_bool('BREVITAS_REINIT_ON_STATE_DICT_LOAD', True)
IGNORE_MISSING_KEYS = env_to_bool('BREVITAS_IGNORE_MISSING_KEYS', False)
# JIT_ENABLED triggers NATIVE_STE_BACKEND_ENABLED to True, but not the other way around
JIT_ENABLED = env_to_bool('BREVITAS_JIT', False) and _enabled
NATIVE_STE_BACKEND_ENABLED = env_to_bool('BREVITAS_NATIVE_STE_BACKEND', False)
VERBOSE = env_to_bool('BREVITAS_VERBOSE', False)

# Internal global variables
_IS_INSIDE_QUANT_LAYER = None
_ONGOING_EXPORT = None



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
import torch.backends.cudnn as cudnn

from brevitas.core.function_wrapper.ops_ste import CeilSte
from brevitas.graph.calibrate import bias_correction_mode
from brevitas.graph.calibrate import calibration_mode
from brevitas.graph.quantize import COMPUTE_LAYER_MAP
from brevitas.graph.quantize import LAYERWISE_COMPUTE_LAYER_MAP
from brevitas.graph.quantize import layerwise_quantize
from brevitas.graph.quantize import QUANT_ACT_MAP
from brevitas.graph.quantize import QUANT_IDENTITY_MAP
from brevitas.graph.quantize import quantize
from brevitas.graph.target.flexml import FLEXML_COMPUTE_LAYER_MAP
from brevitas.graph.target.flexml import FLEXML_QUANT_ACT_MAP
from brevitas.graph.target.flexml import FLEXML_QUANT_IDENTITY_MAP
from brevitas.graph.target.flexml import quantize_flexml
from brevitas.inject.enum import RestrictValueType
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.nn.quant_mha import QuantMultiheadAttention
from brevitas.quant.scaled_int import Int16Bias
from brevitas.quant.scaled_int import Int32Bias
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFixedPoint
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloat

LAYER_MAP = {
    'generic': [COMPUTE_LAYER_MAP, QUANT_ACT_MAP, QUANT_IDENTITY_MAP],
    'layerwise': [LAYERWISE_COMPUTE_LAYER_MAP],
    'flexml': [FLEXML_COMPUTE_LAYER_MAP, FLEXML_QUANT_ACT_MAP, FLEXML_QUANT_IDENTITY_MAP]}

ASYMMETRIC_ACT_QUANT_MAP = {
    'generic': ShiftedUint8ActPerTensorFloat,
    'layerwise': ShiftedUint8ActPerTensorFloat,
    'flexml': ShiftedUint8ActPerTensorFixedPoint}

QUANTIZE_MAP = {'layerwise': layerwise_quantize, 'generic': quantize, 'flexml': quantize_flexml}

BIAS_BIT_WIDTH_MAP = {'int32': Int32Bias, 'int16': Int16Bias}


def quantize_model(
        model,
        backend,
        act_bit_width,
        weight_bit_width,
        bias_bit_width,
        scaling_per_output_channel,
        act_quant_percentile,
        act_quant_type,
        scale_factor_type,
        weight_narrow_range=False):
    # Define what quantize function to use and, based on the given configuration, its arguments
    quantize_fn = QUANTIZE_MAP[backend]

    act_quant_asym = None
    if act_quant_type == 'asymmetric':
        act_quant_asym = ASYMMETRIC_ACT_QUANT_MAP[backend]

    maps = update_quant_maps(
        LAYER_MAP[backend],
        scale_factor_type=scale_factor_type,
        bias_bit_width=bias_bit_width,
        scaling_per_output_channel=scaling_per_output_channel,
        act_quant_percentile=act_quant_percentile,
        act_quant_asym=act_quant_asym,
        act_bit_width=act_bit_width,
        weight_bit_width=weight_bit_width,
        weight_narrow_range=weight_narrow_range)

    if len(maps) == 3:
        # Generic and flexml requires three mappings for quantization
        quantize_kwargs = {
            'compute_layer_map': maps[0], 'quant_act_map': maps[1], 'quant_identity_map': maps[2]}
    elif len(maps) == 1:
        # Layerwise requires only the compute layer mapping
        quantize_kwargs = {'compute_layer_map': maps[0]}

    quant_model = quantize_fn(model, **quantize_kwargs)
    return quant_model


def update_quant_maps(
        maps,
        scale_factor_type,
        bias_bit_width,
        scaling_per_output_channel,
        act_quant_percentile,
        act_quant_asym,
        act_bit_width,
        weight_bit_width,
        weight_narrow_range):
    """
    Starting from pre-defined quantizers, modify them to match the desired configuration
    """

    act_kwargs = {'bit_width': act_bit_width, 'high_percentile_q': act_quant_percentile}

    if act_quant_asym is not None:
        act_kwargs['act_quant'] = act_quant_asym
        act_kwargs['low_percentile_q'] = 100.0 - act_quant_percentile

    weight_kwargs = {
        'scaling_per_output_channel': scaling_per_output_channel,
        'bit_width': weight_bit_width,
        'narrow_range': weight_narrow_range}

    scale_factor_dict = {}
    if scale_factor_type == 'po2':
        scale_factor_dict['restrict_scaling_type'] = RestrictValueType.POWER_OF_TWO
        scale_factor_dict['restrict_value_float_to_int_impl'] = CeilSte
    elif scale_factor_type == 'float32':
        scale_factor_dict['restrict_scaling_type'] = RestrictValueType.FP

    act_kwargs.update(scale_factor_dict)
    weight_kwargs.update(scale_factor_dict)

    def weight_kwargs_prefix(prefix):
        return {prefix + k: v for k, v in weight_kwargs.items()}

    def act_kwargs_prefix(prefix):
        updated_kwargs = {}
        for k, v in act_kwargs.items():
            key = k
            if prefix != '':
                key = prefix + key.replace('act_', '')
            updated_kwargs[key] = v
        return updated_kwargs

    bias_quant = BIAS_BIT_WIDTH_MAP[bias_bit_width]
    for map in maps:
        for k, v in map.items():
            if v is None:
                # Non quantized layer, continue
                continue
            if issubclass(v[0], QuantWBIOL):
                map[k][1].update(weight_kwargs_prefix('weight_'))
                map[k][1]['bias_quant'] = bias_quant
                if act_quant_asym is not None:
                    map[k][1]['return_quant_tensor'] = False
                if 'input_quant' in v[1].keys():
                    # Add kwargs arguments to input_quant, if present
                    map[k][1].update(act_kwargs_prefix('input_'))
            elif v[0] == QuantMultiheadAttention:
                map[k][1].update(weight_kwargs_prefix('in_proj_'))
                map[k][1].update(weight_kwargs_prefix('out_proj_'))
                map[k][1].update(act_kwargs_prefix('attn_output_weights_'))
                map[k][1].update(act_kwargs_prefix('q_scaled_'))
                map[k][1].update(act_kwargs_prefix('k_transposed_'))
                map[k][1].update(act_kwargs_prefix('v_'))
                map[k][1].update(act_kwargs_prefix('out_proj_input_'))
                map[k][1]['in_proj_bias_quant'] = bias_quant
                map[k][1]['out_proj_bias_quant'] = bias_quant
                if act_quant_asym is not None:
                    map[k][1]['return_quant_tensor'] = False
                if 'in_proj_input_quant' in v[1].keys():
                    # Add kwargs arguments to input_quant, if present
                    map[k][1].update(act_kwargs_prefix('in_proj_input_'))
            elif 'act_quant' in v[1].keys():
                # Add kwargs argument to activation quantizers.
                v[1].update(act_kwargs_prefix(''))

    return maps


def calibrate(calib_loader, model, bias_corr=True):
    """
    Perform calibration and bias correction, if enabled
    """
    model.eval()
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    with torch.no_grad():
        with calibration_mode(model):
            for i, (images, target) in enumerate(calib_loader):
                images = images.to(device)
                images = images.to(dtype)
                model(images)

        if bias_corr:
            with bias_correction_mode(model):
                for i, (images, target) in enumerate(calib_loader):
                    images = images.to(device)
                    images = images.to(dtype)
                    model(images)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

from torch import nn

from brevitas.inject.defaults import Int8ActPerTensorFloat
from brevitas.inject.defaults import Int8ActPerTensorFloatMinMaxInit
from brevitas.inject.defaults import Uint8ActPerTensorFloat

from .quant_layer import ActQuantType
from .quant_layer import QuantNonLinearActLayer as QuantNLAL


class QuantReLU(QuantNLAL):

    def __init__(
            self,
            act_quant: Optional[ActQuantType] = Uint8ActPerTensorFloat,
            input_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs):
        QuantNLAL.__init__(
            self,
            act_impl=nn.ReLU,
            passthrough_act=True,
            input_quant=input_quant,
            act_quant=act_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)


class QuantSigmoid(QuantNLAL):

    def __init__(
            self,
            act_quant: Optional[ActQuantType] = Uint8ActPerTensorFloat,
            input_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs):
        QuantNLAL.__init__(
            self,
            act_impl=nn.Sigmoid,
            passthrough_act=False,
            input_quant=input_quant,
            act_quant=act_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)


class QuantTanh(QuantNLAL):

    def __init__(
            self,
            act_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            input_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs):
        QuantNLAL.__init__(
            self,
            act_impl=nn.Tanh,
            passthrough_act=False,
            input_quant=input_quant,
            act_quant=act_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)


class QuantHardTanh(QuantNLAL):

    def __init__(
            self,
            act_quant: Optional[ActQuantType] = Int8ActPerTensorFloatMinMaxInit,
            input_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs):
        QuantNLAL.__init__(
            self,
            act_impl=nn.Hardtanh,
            passthrough_act=True,
            input_quant=input_quant,
            act_quant=act_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)


class QuantIdentity(QuantNLAL):

    def __init__(
            self,
            act_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            return_quant_tensor: bool = False,
            **kwargs):
        QuantNLAL.__init__(
            self,
            input_quant=None,
            act_impl=None,
            passthrough_act=True,
            act_quant=act_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

# retrocompatibility
# the assert prevents the removal of the unused import
from brevitas.inject.enum import QuantType

assert QuantType

from .binary import BinaryQuant
from .binary import ClampedBinaryQuant
from .int import PrescaledRestrictIntQuant
from .int import PrescaledRestrictIntQuantWithInputBitWidth
from .int import RescalingIntQuant
from .int import TruncIntQuant
from .int_base import DecoupledIntQuant
from .int_base import IntQuant
from .ternary import TernaryQuant



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import ast
from functools import reduce
from operator import mul

import torch
from torch.nn import BatchNorm1d
from torch.nn import Dropout
from torch.nn import Module
from torch.nn import ModuleList

from brevitas.nn import QuantIdentity
from brevitas.nn import QuantLinear

from .common import CommonActQuant
from .common import CommonWeightQuant
from .tensor_norm import TensorNorm

DROPOUT = 0.2


class FC(Module):

    def __init__(
        self,
        num_classes,
        weight_bit_width,
        act_bit_width,
        in_bit_width,
        in_channels,
        out_features,
        in_features=(28, 28)):
        super(FC, self).__init__()

        self.features = ModuleList()
        self.features.append(QuantIdentity(act_quant=CommonActQuant, bit_width=in_bit_width))
        self.features.append(Dropout(p=DROPOUT))
        in_features = reduce(mul, in_features)
        for out_features in out_features:
            self.features.append(
                QuantLinear(
                    in_features=in_features,
                    out_features=out_features,
                    bias=False,
                    weight_bit_width=weight_bit_width,
                    weight_quant=CommonWeightQuant))
            in_features = out_features
            self.features.append(BatchNorm1d(num_features=in_features))
            self.features.append(QuantIdentity(act_quant=CommonActQuant, bit_width=act_bit_width))
            self.features.append(Dropout(p=DROPOUT))
        self.features.append(
            QuantLinear(
                in_features=in_features,
                out_features=num_classes,
                bias=False,
                weight_bit_width=weight_bit_width,
                weight_quant=CommonWeightQuant))
        self.features.append(TensorNorm())

        for m in self.modules():
            if isinstance(m, QuantLinear):
                torch.nn.init.uniform_(m.weight.data, -1, 1)

    def clip_weights(self, min_val, max_val):
        for mod in self.features:
            if isinstance(mod, QuantLinear):
                mod.weight.data.clamp_(min_val, max_val)

    def forward(self, x):
        x = x.view(x.shape[0], -1)
        x = 2.0 * x - torch.tensor([1.0], device=x.device)
        for mod in self.features:
            x = mod(x)
        return x


def fc(cfg):
    weight_bit_width = cfg.getint('QUANT', 'WEIGHT_BIT_WIDTH')
    act_bit_width = cfg.getint('QUANT', 'ACT_BIT_WIDTH')
    in_bit_width = cfg.getint('QUANT', 'IN_BIT_WIDTH')
    num_classes = cfg.getint('MODEL', 'NUM_CLASSES')
    in_channels = cfg.getint('MODEL', 'IN_CHANNELS')
    out_features = ast.literal_eval(cfg.get('MODEL', 'OUT_FEATURES'))
    net = FC(
        weight_bit_width=weight_bit_width,
        act_bit_width=act_bit_width,
        in_bit_width=in_bit_width,
        in_channels=in_channels,
        out_features=out_features,
        num_classes=num_classes)
    return net



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn import Parameter

import brevitas
import brevitas.config as config
from brevitas.core.bit_width import BitWidthStatefulConst
from brevitas.core.restrict_val import _RestrictClampValue
from brevitas.core.stats import SCALAR_SHAPE
from brevitas.core.stats.stats_wrapper import _Stats
from brevitas.function import abs_binary_sign_grad

__all__ = [
    "ParameterPreScalingWeightNorm",]


class ParameterPreScalingWeightNorm(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of learned pre-clipping scaling factor to support weight
    normalization-based quantization as proposed in `Quantized Neural Networks for Low-
    Precision Accumulation with Guaranteed Overflow Avoidance` by I. Colbert, A. Pappalardo,
    and J. Petri-Koenig.

    The module parameterizes the pre-clipping scaling factor (i.e., `pre_scale`) of the
    decoupled tensor quantizer (i.e., `DecoupledRescalingIntQuant`) by combining the
    calculuated weight norm stats (i.e., `d_w`) with both the parameterized weight norm
    vector (i.e., `g`) and the post-clipping scaling factor (i.e., `post_scale`). The
    arithmetic is outlined below.

    The formulation for weight normalization-based quantization is given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`
    which we re-write as:
        `y = clip(round(w / pre_scale)) * post_scale`
    where `pre_scale = s * norm(w) / g` and `post_scale = s`.

    Here, `pre_scale` refers to the pre-clipping scaling factor and `post_scale` refers to
    the post-clipping scaling factor.

    Args:
        scaling_impl (Module): post-clipping scaling factor.
        normalize_stats_impl (Module): calculate statistics for normalizing weight parameter.
        scaling_stats_input_view_shape_impl (Module): transforming scaling to a new shape.
        tracked_parameter_list (List[torch.nn.Parameter]): list of tracked weight parameters
            for tensor quantizer.
        pre_scaling_shape (Tuple[int]): shape of pre-clipping scaling factor. Default: None.
        restrict_pre_scaling_impl (Module): restrict pre_scaling_init according to some
            criteria. Default: None.
        pre_scaling_min_val (float): force a lower-bound on scaling_init. Default: None.

    Returns:
        Tensor: scaling factor wrapped in a float torch.Tensor.
    """

    def __init__(
            self,
            scaling_impl: Module,
            normalize_stats_impl: Module,
            scaling_stats_input_view_shape_impl: Module,
            tracked_parameter_list: List[torch.nn.Parameter],
            pre_scaling_shape: Optional[Tuple[int, ...]] = None,
            restrict_pre_scaling_impl: Optional[Module] = None,
            pre_scaling_min_val: Optional[float] = None) -> None:
        super(ParameterPreScalingWeightNorm, self).__init__()

        self.stats = _Stats(normalize_stats_impl, pre_scaling_shape)
        self.stats_input_view_shape_impl = scaling_stats_input_view_shape_impl
        self.scaling_impl = scaling_impl  # this is the post-clipping scaling factor

        if len(tracked_parameter_list) > 1:
            raise NotImplementedError(
                "Error: pre-clipping scales do not currently support multiple tracked quantizers.")
        assert len(tracked_parameter_list) == 1

        # Initialize the weight norm parameter vector from the tracked parameter itself
        param = tracked_parameter_list[0]
        param = self.stats_input_view_shape_impl(param)
        pre_scaling_init = self.stats(param)
        if restrict_pre_scaling_impl is not None:
            pre_scaling_init = restrict_pre_scaling_impl.restrict_init_tensor(pre_scaling_init)
        if pre_scaling_init.shape == SCALAR_SHAPE and pre_scaling_shape is not None:
            pre_scaling_init = torch.full(pre_scaling_shape, pre_scaling_init)
        self.value = Parameter(pre_scaling_init)
        self.restrict_clamp_scaling = _RestrictClampValue(
            pre_scaling_min_val, restrict_pre_scaling_impl)

    @brevitas.jit.script_method
    def forward(self, weights: Tensor) -> Tensor:
        """Takes weights as input and returns the pre-clipping scaling factor"""
        weights = self.stats_input_view_shape_impl(weights)
        d_w = self.stats(weights)  # denominator for weight normalization
        g = abs_binary_sign_grad(self.restrict_clamp_scaling(self.value))  # g
        s = self.scaling_impl(weights)  # s
        value = (s * d_w) / g
        return value

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        value_key = prefix + 'value'
        retrocomp_value_key = prefix + 'learned_value'
        if retrocomp_value_key in state_dict:  # retrocompatibility
            state_dict[value_key] = state_dict.pop(retrocomp_value_key)
        super(ParameterPreScalingWeightNorm, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)


class AccumulatorAwareParameterPreScaling(ParameterPreScalingWeightNorm):
    """
    ScriptModule implementation of learned pre-clipping scaling factor to support
    accumulator-aware quantizaion (A2Q) as proposed in `Quantized Neural Networks for Low-
    Precision Accumulation with Guaranteed Overflow Avoidance` by I. Colbert, A. Pappalardo,
    and J. Petri-Koenig.

    The module parameterizes the pre-clipping scaling factor (i.e., `pre_scale`) of the
    decoupled tensor quantizer (i.e., `DecoupledRescalingIntQuant`) by combining the
    calculated weight norm stats (i.e., `d_w`) with both the parameterized weight norm
    vector (i.e., `g`) and the post-clipping scaling factor (i.e., `post_scale`). The
    parameterized weight norm vector is then clamped to the upper bound on the l1-norm
    of the weights according to the accumulator bit-width bounds derived in the referenced
    paper. The pre-clipping scaling factor arithmetic is outlined below.

    The formulation for weight normalization-based quantization is given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`
    which we re-write as:
        `y = clip(round(w / pre_scale)) * post_scale`
    where `pre_scale = s * norm(w) / g` and `post_scale = s`.

    Here, `pre_scale` refers to the pre-clipping scaling factor and `post_scale` refers to
    the post-clipping scaling factor.

    Args:
        scaling_impl (Module): post-clipping scaling factor.
        normalize_stats_impl (Module): calculate statistics for normalizing weight parameter.
        accumulator_bit_width_impl (Module): module that returns the accumulator bit-width.
        scaling_stats_input_view_shape_impl (Module): transforming scaling to a new shape.
        tracked_parameter_list (List[torch.nn.Parameter]): list of tracked weight parameters
            for tensor quantizer.
        pre_scaling_shape (Tuple[int]): shape of pre-clipping scaling factor. Default: None.
        restrict_pre_scaling_impl (Module): restrict pre_scaling_init according to some
            criteria. Default: None.
        pre_scaling_min_val (float): force a lower-bound on scaling_init. Default: None.

    Returns:
        Tensor: scaling factor wrapped in a float torch.Tensor.
    """

    def __init__(
        self,
        scaling_impl: Module,
        normalize_stats_impl: Module,
        accumulator_bit_width_impl: Module,
        scaling_stats_input_view_shape_impl: Module,
        tracked_parameter_list: List[torch.nn.Parameter],
        pre_scaling_shape: Optional[Tuple[int, ...]] = None,
        restrict_pre_scaling_impl: Optional[Module] = None,
        pre_scaling_min_val: Optional[float] = None,
    ) -> None:
        super().__init__(
            scaling_impl,
            normalize_stats_impl,
            scaling_stats_input_view_shape_impl,
            tracked_parameter_list,
            pre_scaling_shape,
            restrict_pre_scaling_impl,
            pre_scaling_min_val,
        )
        self.accumulator_bit_width = accumulator_bit_width_impl

    @brevitas.jit.script_method
    def get_upper_bound_on_l1_norm(self, input_bit_width: Tensor, input_is_signed: bool) -> Tensor:
        """Calculate the upper bound on the l1-norm of the weights using the derivations from
        `Quantized Neural Networks for Low-Precision Accumulation with Guaranteed Overflow Avoidance`
        by I.Colbert, A.Pappalardo, and J.Petri-Koenig."""
        assert input_bit_width is not None, "A2Q relies on input bit-width."
        assert input_is_signed is not None, "A2Q relies on input sign."
        input_is_signed = float(input_is_signed)  # 1. if signed else 0.
        # This is the minimum of the two maximum magnitudes that P could take, which are -2^{P-1}
        # and 2^{P-1}-1. Note that evaluating to -2^{P-1} would mean there is a possibility of overflow
        # on the positive side of this range.
        max_accumulator_bit_width = self.accumulator_bit_width()  # P
        max_accumulator_mag = pow(2., max_accumulator_bit_width - 1.) - 1.  # 2^{P-1}-1
        # This is the maximum possible magnitude that the input data could take. When the data is signed,
        # this is 2^{N-1}. When the data is unsigned, this is 2^N - 1. We use a slightly looser bound here
        # to simplify our derivations on the export validation.
        max_input_mag_inverse = pow(2., input_is_signed - input_bit_width)
        return max_accumulator_mag * max_input_mag_inverse

    @brevitas.jit.script_method
    def forward(self, weights: Tensor, input_bit_width: Tensor, input_is_signed: bool) -> Tensor:
        """Takes weights as input and returns the pre-clipping scaling factor"""
        weights = self.stats_input_view_shape_impl(weights)
        d_w = self.stats(weights)  # denominator for weight normalization
        s = self.scaling_impl(weights)  # s
        g = abs_binary_sign_grad(self.restrict_clamp_scaling(self.value))  # g
        T = self.get_upper_bound_on_l1_norm(input_bit_width, input_is_signed)  # T / s
        g = torch.clamp_max(g / s, T)
        value = d_w / g  # calculating final pre-clipping scaling factor
        return value



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.quant import TruncIntQuant
from brevitas.inject import ExtendedInjector
from brevitas.inject import value
from brevitas.inject.enum import QuantType
from brevitas.proxy import TruncQuantProxyFromInjector
from brevitas.quant.solver.common import SolveBitWidthImplFromEnum
from brevitas.quant.solver.common import SolveTensorQuantFloatToIntImplFromEnum


class SolveTruncTensorQuantFromEnum(ExtendedInjector):

    @value
    def tensor_quant(quant_type):
        if quant_type == QuantType.FP:
            return None
        elif quant_type == QuantType.INT:
            return TruncIntQuant
        elif quant_type == QuantType.TERNARY:
            raise RuntimeError(f'{quant_type} not supported for truncation.')
        elif quant_type == QuantType.BINARY:
            raise RuntimeError(f'{quant_type} not supported for truncation.')
        else:
            raise RuntimeError(f'{quant_type} not recognized.')


class TruncQuantSolver(SolveBitWidthImplFromEnum,
                       SolveTensorQuantFloatToIntImplFromEnum,
                       SolveTruncTensorQuantFromEnum):
    """
    Translate enum directives to truncation-specific quantization core modules.
    It should be placed last in the list of classes a quantizer inherits from,
    to make sure overrides are correctly captured.
    """
    proxy_class = TruncQuantProxyFromInjector



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.quant.base import *
from brevitas.quant.solver.act import ActQuantSolver
from brevitas.quant.solver.bias import BiasQuantSolver
from brevitas.quant.solver.trunc import TruncQuantSolver
from brevitas.quant.solver.weight import WeightQuantSolver

__all__ = [
    'ShiftedUint8ActPerTensorFixedPoint',
    'ShiftedUint8ActPerTensorFloat',
    'ShiftedUint8WeightPerTensorFloat',
    'ShiftedUint8WeightPerChannelFloat']


class ShiftedUint8ActPerTensorFixedPoint(ShiftedParamFromPercentileUintQuant,
                                         ParamFromRuntimePercentileIntervalScaling,
                                         PerTensorPoTScaling8bit,
                                         ActQuantSolver):
    """
    8-bit per-tensor unsigned int fixed-point activations quantizer with
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    runtime statistics.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFixedPoint)
    """
    pass


class ShiftedUint8ActPerTensorFloat(ShiftedParamFromPercentileUintQuant,
                                    ParamFromRuntimePercentileIntervalScaling,
                                    PerTensorFloatScaling8bit,
                                    ActQuantSolver):
    """
    8-bit per-tensor unsigned int activations quantizer with floating-point scale factor and
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    runtime statistics.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFloat)
    """
    pass


class ShiftedUint8WeightPerTensorFloat(ShiftedMinUintQuant,
                                       MinMaxStatsScaling,
                                       PerTensorFloatScaling8bit,
                                       WeightQuantSolver):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-tensor scale factor and integer
    zero point. Both zero-point and scale factors are based on backpropagated statistics of the
    weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerTensorFloat)
    """
    pass


class ShiftedUint8WeightPerChannelFloat(ShiftedMinUintQuant,
                                        MinMaxStatsScaling,
                                        PerChannelFloatScaling8bit,
                                        WeightQuantSolver):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point. Both zero-point and scale factors are based on backpropagated statistics of the
    weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerChannelFloat)
    """
    pass



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

"""
ScriptModule wrappers for various variants of clamping.
"""

import torch
from torch import Tensor

import brevitas
from brevitas.function import tensor_clamp


class TensorClamp(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~brevitas.function.ops.tensor_clamp`.

    Examples:
        >>> tensor_clamp = TensorClamp()
        >>> min_val = torch.tensor(-2.0)
        >>> max_val = torch.tensor(2.0)
        >>> tensor_clamp(torch.tensor([-3.0, 3.0]), min_val, max_val)
        tensor([-2.,  2.])
    """

    def __init__(self) -> None:
        super(TensorClamp, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: Tensor, min_val: Tensor, max_val: Tensor):
        return tensor_clamp(x, min_val=min_val, max_val=max_val)


class ScalarClamp(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.clamp`.

    Examples:
        >>> scalar_clamp = ScalarClamp(min_val=-2.0, max_val=2.0)
        >>> scalar_clamp(torch.tensor([-3.0, 3.0]))
        tensor([-2.,  2.])
    """

    __constants__ = ['min_val', 'max_val']

    def __init__(self, min_val, max_val) -> None:
        super(ScalarClamp, self).__init__()
        self.min_val = min_val
        self.max_val = max_val

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return torch.clamp(x, min=self.min_val, max=self.max_val)


class ClampMin(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.clamp_min`.

    Examples:
        >>> clamp_min = ClampMin(min_val=-2.0)
        >>> clamp_min(torch.tensor(-3.0))
        tensor(-2.)
    """

    __constants__ = ['min_val']

    def __init__(self, min_val: float) -> None:
        super(ClampMin, self).__init__()
        self.min_val = min_val

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return x.clamp_min(self.min_val)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .base import *
from .calibrate import *
from .equalize import *
from .fixed_point import *
from .per_input import *
from .standardize import *



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from packaging import version
import torch

from brevitas import torch_version

if torch_version < version.parse('1.7'):
    from .backport.torch_function import get_testing_overrides
    from .backport.torch_function._overrides import is_tensor_method_or_property
else:
    from torch.overrides import get_testing_overrides
    from torch.overrides import is_tensor_method_or_property

if torch_version < version.parse('1.8.1'):
    from .backport.graph import Graph
    from .backport.graph import magic_methods
    from .backport.graph import reflectable_magic_methods
    from .backport.graph import Target
    from .backport.graph_module import GraphModule
    from .backport.immutable_collections import immutable_dict
    from .backport.immutable_collections import immutable_list
    from .backport.node import map_arg
    from .backport.proxy import base_types
    from .backport.proxy import Node
    from .backport.proxy import Proxy
    from .backport.symbolic_trace import _autowrap_check
    from .backport.symbolic_trace import _find_proxy
    from .backport.symbolic_trace import _orig_module_call
    from .backport.symbolic_trace import _orig_module_getattr
    from .backport.symbolic_trace import _patch_function
    from .backport.symbolic_trace import _Patcher
    from .backport.symbolic_trace import _wrapped_fns_to_patch
    from .backport.symbolic_trace import _wrapped_methods_to_patch
    from .backport.symbolic_trace import HAS_VARSTUFF
    from .backport.symbolic_trace import map_aggregate
    from .backport.symbolic_trace import Tracer
else:
    from torch.fx import Graph
    from torch.fx import GraphModule
    from torch.fx import map_arg
    from torch.fx import Node
    from torch.fx import Proxy
    from torch.fx import Tracer
    from torch.fx.graph import magic_methods
    from torch.fx.graph import reflectable_magic_methods
    from torch.fx.graph import Target
    from torch.fx.proxy import base_types
    try:
        from torch.fx.immutable_collections import immutable_dict
        from torch.fx.immutable_collections import immutable_list
        from torch.fx.symbolic_trace import _autowrap_check
        from torch.fx.symbolic_trace import _find_proxy
        from torch.fx.symbolic_trace import _orig_module_call
        from torch.fx.symbolic_trace import _orig_module_getattr
        from torch.fx.symbolic_trace import _patch_function
        from torch.fx.symbolic_trace import _Patcher
        from torch.fx.symbolic_trace import _wrapped_fns_to_patch
        from torch.fx.symbolic_trace import _wrapped_methods_to_patch
        from torch.fx.symbolic_trace import HAS_VARSTUFF
        from torch.fx.symbolic_trace import map_aggregate
    except ImportError:
        from torch.fx._symbolic_trace import _orig_module_call
        from torch.fx._symbolic_trace import _orig_module_getattr
        from torch.fx._symbolic_trace import _autowrap_check
        from torch.fx._symbolic_trace import _Patcher, map_aggregate
        from torch.fx._symbolic_trace import _wrapped_fns_to_patch, _wrapped_methods_to_patch
        from torch.fx._symbolic_trace import _find_proxy, _patch_function, HAS_VARSTUFF
        from torch.fx.immutable_collections import immutable_dict, immutable_list

from .brevitas_tracer import brevitas_symbolic_trace
from .brevitas_tracer import brevitas_value_trace
from .brevitas_tracer import symbolic_trace
from .brevitas_tracer import value_trace



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import List, Optional, Tuple

import torch
from torch import Tensor
from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from brevitas.function import max_int
from brevitas.quant_tensor import QuantTensor

from .quant_proxy import QuantProxyFromInjector
from .quant_proxy import QuantProxyProtocol

__all__ = [
    'WeightQuantProxyFromInjector',
    'BiasQuantProxyFromInjector',
    'WeightQuantProxyProtocol',
    'BiasQuantProxyProtocol']


@runtime_checkable
class WeightQuantProxyProtocol(QuantProxyProtocol, Protocol):

    def forward(self, x: torch.Tensor) -> QuantTensor:
        ...


@runtime_checkable
class BiasQuantProxyProtocol(QuantProxyProtocol, Protocol):
    requires_input_bit_width: bool
    requires_input_scale: bool

    def forward(
            self, x: Tensor, input_scale: Optional[Tensor],
            input_bit_width: Optional[Tensor]) -> QuantTensor:
        ...


class ParameterQuantProxyFromInjector(QuantProxyFromInjector):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def tracked_parameter_list(self):
        pass

    def init_tensor_quant(self):
        param_list = self.tracked_parameter_list
        # params might not be there yet, e.g. bias before merging
        if param_list:
            self.quant_injector = self.quant_injector.let(tracked_parameter_list=param_list)
            super(ParameterQuantProxyFromInjector, self).init_tensor_quant()

    def max_uint_value(self, bit_width):
        return max_int(False, self.is_narrow_range, bit_width)


class WeightQuantProxyFromInjector(ParameterQuantProxyFromInjector, WeightQuantProxyProtocol):

    @property
    def tracked_parameter_list(self):
        return [m.weight for m in self.tracked_module_list if m.weight is not None]

    @property
    def requires_quant_input(self):
        return False

    def scale(self):
        scale = self.__call__(self.tracked_parameter_list[0]).scale
        return scale

    def zero_point(self):
        zero_point = self.__call__(self.tracked_parameter_list[0]).zero_point
        return zero_point

    def bit_width(self):
        bit_width_ = self.__call__(self.tracked_parameter_list[0]).bit_width
        return bit_width_

    def forward(self, x: torch.Tensor) -> QuantTensor:
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            out, scale, zero_point, bit_width = impl(x)
            return QuantTensor(out, scale, zero_point, bit_width, self.is_signed, self.training)
        else:  # quantization disabled
            return QuantTensor(x, training=self.training)


class DecoupledWeightQuantProxyFromInjector(WeightQuantProxyFromInjector):

    def pre_scale(self):
        output_tuple = self.tensor_quant(self.tracked_parameter_list[0])
        out, scale, zero_point, bit_width, pre_scale, pre_zero_point = output_tuple
        return pre_scale

    def pre_zero_point(self):
        output_tuple = self.tensor_quant(self.tracked_parameter_list[0])
        out, scale, zero_point, bit_width, pre_scale, pre_zero_point = output_tuple
        return pre_zero_point

    def forward(self, x: torch.Tensor) -> QuantTensor:
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            out, scale, zero_point, bit_width, pre_scale, pre_zero_point = impl(x)
            return QuantTensor(out, scale, zero_point, bit_width, self.is_signed, self.training)
        else:  # quantization disabled
            return QuantTensor(x, training=self.training)


class DecoupledWeightQuantWithInputProxyFromInjector(DecoupledWeightQuantProxyFromInjector):

    @property
    def requires_quant_input(self):
        return True

    def scale(self):
        raise NotImplementedError

    def zero_point(self):
        raise NotImplementedError

    def bit_width(self):
        raise NotImplementedError

    def pre_scale(self):
        raise NotImplementedError

    def pre_zero_point(self):
        raise NotImplementedError

    def forward(
            self, x: torch.Tensor, input_bit_width: torch.Tensor,
            input_is_signed: bool) -> QuantTensor:
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            out, scale, zero_point, bit_width, pre_scale, pre_zero_point = impl(x, input_bit_width, input_is_signed)
            return QuantTensor(out, scale, zero_point, bit_width, self.is_signed, self.training)
        else:  # quantization disabled
            return QuantTensor(x, training=self.training)


class BiasQuantProxyFromInjector(ParameterQuantProxyFromInjector, BiasQuantProxyProtocol):

    @property
    def tracked_parameter_list(self):
        return [m.bias for m in self.tracked_module_list if m.bias is not None]

    @property
    def requires_input_bit_width(self) -> bool:
        if self.is_quant_enabled:
            return self.quant_injector.requires_input_bit_width
        else:
            return False

    @property
    def requires_input_scale(self) -> bool:
        if self.is_quant_enabled:
            return self.quant_injector.requires_input_scale
        else:
            return False

    def scale(self):
        if self.requires_input_scale:
            return None
        zhs = self._zero_hw_sentinel()
        scale = self.__call__(self.tracked_parameter_list[0], zhs, zhs).scale
        return scale

    def zero_point(self):
        zhs = self._zero_hw_sentinel()
        zero_point = self.__call__(self.tracked_parameter_list[0], zhs, zhs).zero_point
        return zero_point

    def bit_width(self):
        if self.requires_input_bit_width:
            return None
        zhs = self._zero_hw_sentinel()
        bit_width = self.__call__(self.tracked_parameter_list[0], zhs, zhs).bit_width
        return bit_width

    def forward(
            self,
            x: Tensor,
            input_scale: Optional[Tensor] = None,
            input_bit_width: Optional[Tensor] = None) -> QuantTensor:
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            if self.requires_input_scale and input_scale is None:
                raise RuntimeError("Input scale required")
            if self.requires_input_bit_width and input_bit_width is None:
                raise RuntimeError("Input bit-width required")
            if self.requires_input_scale and self.requires_input_bit_width:
                input_scale = input_scale.view(-1)
                out, out_scale, out_zp, out_bit_width = impl(x, input_scale, input_bit_width)
            elif self.requires_input_scale and not self.requires_input_bit_width:
                input_scale = input_scale.view(-1)
                out, out_scale, out_zp, out_bit_width = impl(x, input_scale)
            elif not self.requires_input_scale and not self.requires_input_bit_width:
                out, out_scale, out_zp, out_bit_width = impl(x)
            else:
                raise RuntimeError("Internally defined bit-width required")
            return QuantTensor(out, out_scale, out_zp, out_bit_width, self.is_signed, self.training)
        else:
            return QuantTensor(x, training=self.training)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from functools import partial
import math
from typing import List, Optional, Tuple

import torch
from torch import nn
from torch import Tensor
import torch.nn.functional as F

import brevitas
from brevitas.nn import QuantIdentity
from brevitas.nn import QuantSigmoid
from brevitas.nn import QuantTanh
from brevitas.nn.mixin import QuantBiasMixin
from brevitas.nn.mixin import QuantWeightMixin
from brevitas.nn.mixin.base import QuantRecurrentLayerMixin
from brevitas.quant import Int8ActPerTensorFloat
from brevitas.quant import Int8WeightPerTensorFloat
from brevitas.quant import Int32Bias
from brevitas.quant import Uint8ActPerTensorFloat
from brevitas.quant_tensor import QuantTensor

QuantTupleShortEnabled = List[Tuple[Tensor, Tensor, Tensor, Tensor]]
QuantTupleShortDisabled = List[Tuple[Tensor, Optional[Tensor], Optional[Tensor], Optional[Tensor]]]
QuantTupleLongEnabled = List[Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]]
QuantTupleLongDisabled = List[Tuple[Tensor,
                                    Optional[Tensor],
                                    Optional[Tensor],
                                    Optional[Tensor],
                                    Optional[Tensor],
                                    Optional[Tensor]]]


class GateWeight(QuantWeightMixin, nn.Module):

    def __init__(self, input_features, output_features, weight_quant, **kwargs):
        nn.Module.__init__(self)
        self.weight = nn.Parameter(torch.randn(output_features, input_features))
        QuantWeightMixin.__init__(self, weight_quant=weight_quant, **kwargs)

    @property
    def output_channel_dim(self):
        return 0

    @property
    def out_channels(self):
        return self.weight.size(self.output_channel_dim)

    def forward(self):
        return self.weight_quant(self.weight)


class GateParams(QuantBiasMixin, nn.Module):

    def __init__(
            self, input_size, hidden_size, bias, weight_quant, bias_quant, input_weight, **kwargs):
        nn.Module.__init__(self)
        if bias:
            self.bias = nn.Parameter(torch.randn(hidden_size))
        else:
            self.bias = None
        QuantBiasMixin.__init__(self, bias_quant, **kwargs)
        if input_weight is None:
            input_weight = GateWeight(input_size, hidden_size, weight_quant=weight_quant, **kwargs)
        self.input_weight = input_weight
        # The quantizer is shared among input-to-hidden and hidden-to-hidden weights
        self.hidden_weight = GateWeight(
            hidden_size, hidden_size, weight_quant=input_weight.weight_quant)


# Simply putting the forward in a standalone script function is not enough
# for the compiler to realize that the bools being passed in are all constants
# so we resort to a module instead
class _QuantStatesInit(nn.Module):
    __constants__ = ['fast_impl', 'quant_enabled']

    def __init__(self, fast_impl: bool, quant_enabled: bool):
        super(_QuantStatesInit, self).__init__()
        self.fast_impl = fast_impl
        self.quant_enabled = quant_enabled

    def forward(self):
        if self.fast_impl:
            if self.quant_enabled:
                quant_states = torch.jit.annotate(QuantTupleShortEnabled, [])
            else:
                quant_states = torch.jit.annotate(QuantTupleShortDisabled, [])
        else:
            if self.quant_enabled:
                quant_states = torch.jit.annotate(QuantTupleLongEnabled, [])
            else:
                quant_states = torch.jit.annotate(QuantTupleLongDisabled, [])
        return quant_states


class _QuantRNNCell(nn.Module):
    __constants__ = ['reverse_input', 'batch_first']

    def __init__(
            self,
            act_fn: nn.Module,
            gate_acc_quant: nn.Module,
            output_quant: nn.Module,
            reverse_input: bool,
            batch_first: bool,
            output_quant_enabled: bool,
            fast_impl: bool):
        super(_QuantRNNCell, self).__init__()
        self.act_fn = act_fn
        self.gate_acc_quant = gate_acc_quant
        self.output_quant = output_quant
        self.reverse_input = reverse_input
        self.batch_first = batch_first
        self.hidden_states_init = _QuantStatesInit(fast_impl, output_quant_enabled)

    def forward_iter(self, quant_input, quant_state, quant_weight_ih, quant_weight_hh, quant_bias):
        quant_gate_ih = F.linear(quant_input, quant_weight_ih)
        quant_gate_hh = F.linear(quant_state, quant_weight_hh)
        quant_gate = self.gate_acc_quant(quant_gate_ih + quant_gate_hh + quant_bias)[0]
        quant_gate = self.act_fn(quant_gate)
        quant_state_tuple = self.output_quant(quant_gate)
        return quant_state_tuple

    def forward(
            self,
            quant_input: Tensor,
            quant_state: Tensor,
            quant_weight_ih: Tensor,
            quant_weight_hh: Tensor,
            quant_bias: Tensor):
        if self.batch_first:
            quant_inputs = quant_input.unbind(1)
        else:
            quant_inputs = quant_input.unbind(0)
        end = len(quant_inputs)
        step = 1
        index = 0
        quant_outputs = self.hidden_states_init()
        if self.reverse_input:
            index = end - 1
            step = -1
        for _ in range(end):
            quant_input = quant_inputs[index]
            quant_state_tuple = self.forward_iter(
                quant_input, quant_state, quant_weight_ih, quant_weight_hh, quant_bias)
            index = index + step
            quant_outputs += [quant_state_tuple]
            quant_state = quant_state_tuple[0]
        return quant_outputs


class _QuantLSTMCell(nn.Module):
    __constants__ = ['reverse_input', 'batch_first', 'cifg']

    def __init__(
            self,
            output_quant,
            cell_state_quant,
            input_acc_quant,
            forget_acc_quant,
            cell_acc_quant,
            output_acc_quant,
            input_sigmoid_quant,
            forget_sigmoid_quant,
            cell_tanh_quant,
            output_sigmoid_quant,
            hidden_state_tanh_quant,
            reverse_input: bool,
            batch_first: bool,
            cifg: bool,
            output_quant_enabled: bool,
            cell_state_quant_enabled: bool,
            fast_impl: bool):
        super(_QuantLSTMCell, self).__init__()
        self.output_quant = output_quant
        self.cell_state_quant = cell_state_quant
        self.input_acc_quant = input_acc_quant
        self.forget_acc_quant = forget_acc_quant
        self.cell_acc_quant = cell_acc_quant
        self.output_acc_quant = output_acc_quant
        self.input_sigmoid_quant = input_sigmoid_quant
        self.forget_sigmoid_quant = forget_sigmoid_quant
        self.cell_tanh_quant = cell_tanh_quant
        self.output_sigmoid_quant = output_sigmoid_quant
        self.hidden_state_tanh_quant = hidden_state_tanh_quant
        self.reverse_input = reverse_input
        self.batch_first = batch_first
        self.cifg = cifg
        self.hidden_states_init = _QuantStatesInit(fast_impl, output_quant_enabled)
        self.cell_states_init = _QuantStatesInit(fast_impl, cell_state_quant_enabled)

    def forward_iter(
            self,
            quant_input: Tensor,
            quant_hidden_state: Tensor,
            quant_cell_state: Tensor,
            quant_weight_ii: Tensor,
            quant_weight_if: Tensor,
            quant_weight_ic: Tensor,
            quant_weight_io: Tensor,
            quant_weight_hi: Tensor,
            quant_weight_hf: Tensor,
            quant_weight_hc: Tensor,
            quant_weight_ho: Tensor,
            quant_bias_input: Tensor,
            quant_bias_forget: Tensor,
            quant_bias_cell: Tensor,
            quant_bias_output: Tensor):
        # Input gate
        quant_ii_gate = F.linear(quant_input, quant_weight_ii)
        quant_hi_gate = F.linear(quant_hidden_state, quant_weight_hi)
        quant_input_gate = self.input_acc_quant(quant_ii_gate + quant_hi_gate + quant_bias_input)[0]
        quant_input_gate = self.input_sigmoid_quant(quant_input_gate)[0]
        # Forget gate
        if self.cifg:
            quant_ones = self.input_sigmoid_quant.tensor_quant(torch.ones_like(quant_input_gate))[0]
            # CIFG is defined as 1 - input_gate, in line with ONNXRuntime
            quant_forget_gate = quant_ones - quant_input_gate
        else:
            quant_if_gate = F.linear(quant_input, quant_weight_if)
            quant_hf_gate = F.linear(quant_hidden_state, quant_weight_hf)
            quant_forget_gate = self.forget_acc_quant(
                quant_if_gate + quant_hf_gate + quant_bias_forget)[0]
            quant_forget_gate = self.forget_sigmoid_quant(quant_forget_gate)[0]
        # Cell gate
        quant_ic_gate = F.linear(quant_input, quant_weight_ic)
        quant_hc_gate = F.linear(quant_hidden_state, quant_weight_hc)
        quant_cell_gate = self.cell_acc_quant(quant_ic_gate + quant_hc_gate + quant_bias_cell)[0]
        quant_cell_gate = self.cell_tanh_quant(quant_cell_gate)[0]
        # Output gate
        quant_io_gate = F.linear(quant_input, quant_weight_io)
        quant_ho_gate = F.linear(quant_hidden_state, quant_weight_ho)
        quant_out_gate = self.output_acc_quant(quant_io_gate + quant_ho_gate + quant_bias_output)[0]
        quant_out_gate = self.output_sigmoid_quant(quant_out_gate)[0]
        quant_forget_cell = self.cell_state_quant(quant_forget_gate * quant_cell_state)[0]
        quant_inp_cell = self.cell_state_quant(quant_input_gate * quant_cell_gate)[0]
        quant_cell_state_tuple = self.cell_state_quant(quant_forget_cell + quant_inp_cell)
        quant_hidden_state_tanh = self.hidden_state_tanh_quant(quant_cell_state_tuple[0])[0]
        quant_hidden_state = quant_out_gate * quant_hidden_state_tanh
        quant_hidden_state_tuple = self.output_quant(quant_hidden_state)
        return quant_hidden_state_tuple, quant_cell_state_tuple

    def forward(
            self,
            quant_input: Tensor,
            quant_hidden_state: Tensor,
            quant_cell_state: Tensor,
            quant_weight_ii: Tensor,
            quant_weight_if: Tensor,
            quant_weight_ic: Tensor,
            quant_weight_io: Tensor,
            quant_weight_hi: Tensor,
            quant_weight_hf: Tensor,
            quant_weight_hc: Tensor,
            quant_weight_ho: Tensor,
            quant_bias_input: Tensor,
            quant_bias_forget: Tensor,
            quant_bias_cell: Tensor,
            quant_bias_output: Tensor):
        if self.batch_first:
            seq_dim = 1
        else:
            seq_dim = 0
        quant_inputs = quant_input.unbind(seq_dim)
        end = len(quant_inputs)
        step = 1
        index = 0
        if self.reverse_input:
            index = end - 1
            step = -1
        quant_hidden_states = self.hidden_states_init()
        quant_cell_states = self.cell_states_init()
        for _ in range(end):
            quant_input = quant_inputs[index]
            quant_hidden_state_tuple, quant_cell_state_tuple = self.forward_iter(
                quant_input,
                quant_hidden_state,
                quant_cell_state,
                quant_weight_ii,
                quant_weight_if,
                quant_weight_ic,
                quant_weight_io,
                quant_weight_hi,
                quant_weight_hf,
                quant_weight_hc,
                quant_weight_ho,
                quant_bias_input,
                quant_bias_forget,
                quant_bias_cell,
                quant_bias_output)
            index = index + step
            quant_hidden_states += [quant_hidden_state_tuple]
            quant_hidden_state = quant_hidden_state_tuple[0]
            quant_cell_states += [quant_cell_state_tuple]
            quant_cell_state = quant_cell_state_tuple[0]
        return quant_hidden_states, quant_hidden_states[-1], quant_cell_states[-1]


class _QuantRNNLayer(QuantRecurrentLayerMixin, nn.Module):

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            batch_first: bool,
            weight_quant,
            bias_quant,
            io_quant,
            gate_acc_quant,
            reverse_input: bool,
            quantize_output_only: bool,
            shared_input_hidden_weights: bool,
            return_quant_tensor: bool,
            nonlinearity: str,
            input_weight: GateWeight = None,
            **kwargs):
        nn.Module.__init__(self)
        io_quant = QuantIdentity(io_quant, act_kwargs_prefix='io_', **kwargs)
        gate_acc_quant = QuantIdentity(
            gate_acc_quant, act_kwargs_prefix='gate_acc_quant_', **kwargs)
        if nonlinearity == 'tanh':
            act_fn = nn.Tanh()
        elif nonlinearity == 'relu':
            act_fn = nn.ReLU()
        else:
            raise RuntimeError(f"{nonlinearity} not supported.")
        cell = _QuantRNNCell(
            act_fn,
            gate_acc_quant.act_quant,
            io_quant.act_quant,
            reverse_input,
            batch_first,
            io_quant.act_quant.is_quant_enabled,
            fast_impl=False)
        QuantRecurrentLayerMixin.__init__(
            self,
            cell=cell,
            io_quant=io_quant.act_quant,
            input_size=input_size,
            hidden_size=hidden_size,
            reverse_input=reverse_input,
            quantize_output_only=quantize_output_only,
            shared_input_hidden_weights=shared_input_hidden_weights,
            return_quant_tensor=return_quant_tensor)
        self.gate_params = GateParams(
            input_size, hidden_size, bias, weight_quant, bias_quant, input_weight, **kwargs)
        self.reset_parameters()

    @property
    def weights_to_share(self):
        if self.shared_input_hidden_weights:
            return {'input_weight': self.gate_params.input_weight}
        else:
            return {}

    @property
    def quantizers_to_share(self):
        shared_quantizers = {'io_quant': self.io_quant}
        return shared_quantizers

    @property
    def fast_cell(self):
        if self._fast_cell is not None:
            return self._fast_cell
        else:
            # lazy late init to make sure the correct fused act quant proxy is captured
            # since on every sharing among layers at init time it is re-initialized
            self._fast_cell = _QuantRNNCell(
                self.cell.act_fn,
                self._wrap_act_proxy('gate_acc_quant'),
                self._wrap_act_proxy('output_quant'),
                self.cell.reverse_input,
                self.cell.batch_first,
                self.cell.output_quant.is_quant_enabled,
                fast_impl=True)
            if brevitas.config.JIT_ENABLED:
                self._fast_cell = torch.jit.script(self._fast_cell)
            return self._fast_cell

    def forward(self, inp, state):
        quant_input = self.maybe_quantize_input(inp)
        quant_weight_ih, quant_weight_hh, quant_bias = self.gate_params_fwd(
            self.gate_params, quant_input)
        if quant_bias.value is None:
            quant_bias = torch.tensor(0., device=quant_input.value.device)
        else:
            quant_bias = quant_bias.value
        quant_state = self.maybe_quantize_state(quant_input.value, state, self.cell.output_quant)
        if self.export_mode:
            cell = self.export_handler
        elif self.fast_mode:
            cell = self.fast_cell
        else:
            cell = self.cell
        quant_outputs = cell(
            quant_input.value,
            quant_state.value,
            quant_weight_ih.value,
            quant_weight_hh.value,
            quant_bias)
        quant_output = self.pack_quant_outputs(quant_outputs)
        quant_state = self.pack_quant_state(quant_outputs[-1], self.cell.output_quant)
        return quant_output, quant_state

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        hs = self.hidden_size

        def bias():
            bias_name = f'{prefix}gate_params.bias'
            if not bias_name in state_dict.keys():
                state_dict[bias_name] = torch.zeros(hs)
            return state_dict[bias_name]

        def _set_weight(value, input_name):
            key = f'{prefix}gate_params.{input_name}_weight.weight'
            state_dict[key] = value

        set_input_weight = partial(_set_weight, input_name='input')
        set_hidden_weight = partial(_set_weight, input_name='hidden')

        for name, value in list(state_dict.items()):
            if prefix + 'bias_ih' in name or prefix + 'bias_hh' in name:
                bias().add_(value[:hs])
                del state_dict[name]
            elif prefix + 'weight_ih' in name:
                set_input_weight(value[:hs, :])
                del state_dict[name]
            elif prefix + 'weight_hh' in name:
                set_hidden_weight(value[:hs, :])
                del state_dict[name]

        super(_QuantRNNLayer, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


class _QuantLSTMLayer(QuantRecurrentLayerMixin, nn.Module):

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            batch_first: bool,
            weight_quant,
            bias_quant,
            io_quant,
            gate_acc_quant,
            sigmoid_quant,
            tanh_quant,
            cell_state_quant,
            reverse_input: bool,
            quantize_output_only: bool,
            cifg: bool,
            shared_input_hidden_weights: bool,
            shared_cell_state_quant: bool,
            shared_intra_layer_weight_quant: bool,
            shared_intra_layer_gate_acc_quant: bool,
            return_quant_tensor: bool,
            input_input_weight: GateWeight = None,
            input_forget_weight: GateWeight = None,
            input_cell_weight: GateWeight = None,
            input_output_weight: GateWeight = None,
            **kwargs):
        nn.Module.__init__(self)
        io_quant = QuantIdentity(io_quant, act_kwargs_prefix='io_', **kwargs)
        cell_state_quant = QuantIdentity(
            cell_state_quant, act_kwargs_prefix='cell_state_', **kwargs)

        # Quantizers for the output accumulator of each gate
        input_acc_quant = QuantIdentity(gate_acc_quant, act_kwargs_prefix='gate_acc_', **kwargs)
        if shared_intra_layer_gate_acc_quant:
            forget_acc_quant = input_acc_quant
            cell_acc_quant = input_acc_quant
            output_acc_quant = input_acc_quant
        else:
            cell_acc_quant = QuantIdentity(gate_acc_quant, act_kwargs_prefix='gate_acc_', **kwargs)
            output_acc_quant = QuantIdentity(
                gate_acc_quant, act_kwargs_prefix='gate_acc_', **kwargs)
            if cifg:
                # Avoid dealing with None and set it to the input one
                forget_acc_quant = input_acc_quant
            else:
                forget_acc_quant = QuantIdentity(
                    gate_acc_quant, act_kwargs_prefix='gate_acc_', **kwargs)

        # Internal quantized activations
        input_sigmoid_quant = QuantSigmoid(sigmoid_quant, act_kwargs_prefix='sigmoid_', **kwargs)
        cell_tanh_quant = QuantTanh(tanh_quant, act_kwargs_prefix='tanh_', **kwargs)
        output_sigmoid_quant = QuantSigmoid(sigmoid_quant, act_kwargs_prefix='sigmoid_', **kwargs)
        hidden_state_tanh_quant = QuantTanh(tanh_quant, act_kwargs_prefix='tanh_', **kwargs)
        if cifg:
            # Avoid dealing with None and set it to the input one
            forget_sigmoid_quant = input_sigmoid_quant
        else:
            forget_sigmoid_quant = QuantSigmoid(
                sigmoid_quant, act_kwargs_prefix='sigmoid_', **kwargs)

        cell = _QuantLSTMCell(
            output_quant=io_quant.act_quant,
            cell_state_quant=cell_state_quant.act_quant,
            input_acc_quant=input_acc_quant.act_quant,
            forget_acc_quant=forget_acc_quant.act_quant,
            cell_acc_quant=cell_acc_quant.act_quant,
            output_acc_quant=output_acc_quant.act_quant,
            input_sigmoid_quant=input_sigmoid_quant.act_quant,
            forget_sigmoid_quant=forget_sigmoid_quant.act_quant,
            cell_tanh_quant=cell_tanh_quant.act_quant,
            output_sigmoid_quant=output_sigmoid_quant.act_quant,
            hidden_state_tanh_quant=hidden_state_tanh_quant.act_quant,
            reverse_input=reverse_input,
            batch_first=batch_first,
            cifg=cifg,
            output_quant_enabled=io_quant.act_quant.is_quant_enabled,
            cell_state_quant_enabled=cell_state_quant.act_quant.is_quant_enabled,
            fast_impl=False)
        QuantRecurrentLayerMixin.__init__(
            self,
            cell=cell,
            io_quant=io_quant.act_quant,
            input_size=input_size,
            hidden_size=hidden_size,
            reverse_input=reverse_input,
            quantize_output_only=quantize_output_only,
            shared_input_hidden_weights=shared_input_hidden_weights,
            return_quant_tensor=return_quant_tensor)

        self.input_gate_params = GateParams(
            input_size, hidden_size, bias, weight_quant, bias_quant, input_forget_weight, **kwargs)
        if shared_intra_layer_weight_quant:
            # Share the input-to-hidden input weight quantizer, which is also shared with hidden-to-hidden
            weight_quant = self.input_gate_params.input_weight.weight_quant
        if cifg:
            self.forget_gate_params = None
        else:
            self.forget_gate_params = GateParams(
                input_size,
                hidden_size,
                bias,
                weight_quant,
                bias_quant,
                input_input_weight,
                **kwargs)
        self.cell_gate_params = GateParams(
            input_size, hidden_size, bias, weight_quant, bias_quant, input_cell_weight, **kwargs)
        self.output_gate_params = GateParams(
            input_size, hidden_size, bias, weight_quant, bias_quant, input_output_weight, **kwargs)
        self.shared_cell_state_quant = shared_cell_state_quant
        self.cifg = cifg
        self.reset_parameters()

    @property
    def weights_to_share(self):
        if self.shared_input_hidden_weights:
            out = {
                'input_input_weight': self.input_gate_params.input_weight,
                'input_cell_weight': self.cell_gate_params.input_weight,
                'input_output_weight': self.output_gate_params.input_weight}
            if not self.cifg:
                out['input_forget_weight'] = self.forget_gate_params.input_weight
            return out
        else:
            return {}

    @property
    def quantizers_to_share(self):
        shared_quantizers = {'io_quant': self.io_quant}
        if self.shared_cell_state_quant:
            shared_quantizers['cell_state_quant'] = self.cell.cell_state_quant
        return shared_quantizers

    @property
    def fast_cell(self):
        if self._fast_cell is not None:
            return self._fast_cell
        else:
            # lazy late init to make sure the correct fused act quant proxy is captured
            # since on every sharing among layers at init time it is re-initialized
            self._fast_cell = _QuantLSTMCell(
                # Wrap a None or act_fn only proxy to still return a tuple
                # whenever quantization is disabled
                output_quant=self._wrap_act_proxy('output_quant'),
                cell_state_quant=self._wrap_act_proxy('cell_state_quant'),
                input_acc_quant=self._wrap_act_proxy('input_acc_quant'),
                forget_acc_quant=self._wrap_act_proxy('forget_acc_quant'),
                cell_acc_quant=self._wrap_act_proxy('cell_acc_quant'),
                output_acc_quant=self._wrap_act_proxy('output_acc_quant'),
                input_sigmoid_quant=self._wrap_act_proxy('input_sigmoid_quant'),
                forget_sigmoid_quant=self._wrap_act_proxy('forget_sigmoid_quant'),
                cell_tanh_quant=self._wrap_act_proxy('cell_tanh_quant'),
                output_sigmoid_quant=self._wrap_act_proxy('output_sigmoid_quant'),
                hidden_state_tanh_quant=self._wrap_act_proxy('hidden_state_tanh_quant'),
                reverse_input=self.cell.reverse_input,
                batch_first=self.cell.batch_first,
                cifg=self.cell.cifg,
                output_quant_enabled=self.cell.output_quant.is_quant_enabled,
                cell_state_quant_enabled=self.cell.cell_state_quant.is_quant_enabled,
                fast_impl=True)
            if brevitas.config.JIT_ENABLED:
                self._fast_cell = torch.jit.script(self._fast_cell)
            return self._fast_cell

    def forward(self, inp, hidden_state, cell_state):
        quant_input = self.maybe_quantize_input(inp)
        quant_weight_ii, quant_weight_hi, quant_bias_input = self.gate_params_fwd(
            self.input_gate_params, quant_input)
        quant_weight_ic, quant_weight_hc, quant_bias_cell = self.gate_params_fwd(
            self.cell_gate_params, quant_input)
        quant_weight_io, quant_weight_ho, quant_bias_output = self.gate_params_fwd(
            self.output_gate_params, quant_input)
        if self.cifg:
            # Avoid dealing with None and set it the same as the forget one
            quant_weight_if = quant_weight_ii
            quant_weight_hf = quant_weight_hi
            quant_bias_forget = quant_bias_input
        else:
            quant_weight_if, quant_weight_hf, quant_bias_forget = self.gate_params_fwd(
                self.forget_gate_params, quant_input)
        # Handle None bias by setting it 0.
        if quant_bias_input.value is None:
            quant_bias_input = torch.tensor(0., device=quant_input.value.device)
        else:
            quant_bias_input = quant_bias_input.value
        if quant_bias_forget.value is None:
            quant_bias_forget = torch.tensor(0., device=quant_input.value.device)
        else:
            quant_bias_forget = quant_bias_forget.value
        if quant_bias_cell.value is None:
            quant_bias_cell = torch.tensor(0., device=quant_input.value.device)
        else:
            quant_bias_cell = quant_bias_cell.value
        if quant_bias_output.value is None:
            quant_bias_output = torch.tensor(0., device=quant_input.value.device)
        else:
            quant_bias_output = quant_bias_output.value
        quant_hidden_state = self.maybe_quantize_state(
            quant_input.value, hidden_state, self.cell.output_quant)
        quant_cell_state = self.maybe_quantize_state(
            quant_input.value, cell_state, self.cell.cell_state_quant)
        # Pick cell impl
        if self.export_mode:
            cell = self.export_handler
        elif self.fast_mode:
            cell = self.fast_cell
        else:
            cell = self.cell
        quant_outputs, quant_hidden_state, quant_cell_state = cell(
            quant_input.value,
            quant_hidden_state.value,
            quant_cell_state.value,
            quant_weight_ii=quant_weight_ii.value,
            quant_weight_if=quant_weight_if.value,
            quant_weight_ic=quant_weight_ic.value,
            quant_weight_io=quant_weight_io.value,
            quant_weight_hi=quant_weight_hi.value,
            quant_weight_hf=quant_weight_hf.value,
            quant_weight_hc=quant_weight_hc.value,
            quant_weight_ho=quant_weight_ho.value,
            quant_bias_input=quant_bias_input,
            quant_bias_forget=quant_bias_forget,
            quant_bias_cell=quant_bias_cell,
            quant_bias_output=quant_bias_output)
        quant_outputs = self.pack_quant_outputs(quant_outputs)
        quant_hidden_state = self.pack_quant_state(quant_hidden_state, self.cell.output_quant)
        quant_cell_state = self.pack_quant_state(quant_cell_state, self.cell.cell_state_quant)
        return quant_outputs, quant_hidden_state, quant_cell_state

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        hs = self.hidden_size

        def bias(gate_name):
            bias_name = f'{prefix}{gate_name}_gate_params.bias'
            if not bias_name in state_dict.keys():
                state_dict[bias_name] = torch.zeros(hs)
            return state_dict[bias_name]

        def _set_weight(gate_name, value, input_name):
            key = f'{prefix}{gate_name}_gate_params.{input_name}_weight.weight'
            state_dict[key] = value

        set_input_weight = partial(_set_weight, input_name='input')
        set_hidden_weight = partial(_set_weight, input_name='hidden')

        for name, value in list(state_dict.items()):
            if prefix + 'bias_ih' in name or prefix + 'bias_hh' in name:
                bias('input').add_(value[:hs])
                bias('forget').add_(value[hs:hs * 2])
                bias('cell').add_(value[2 * hs:hs * 3])
                bias('output').add_(value[3 * hs:])
                del state_dict[name]
            elif prefix + 'weight_ih' in name:
                set_input_weight('input', value[:hs, :])
                set_input_weight('forget', value[hs:hs * 2, :])
                set_input_weight('cell', value[2 * hs:hs * 3, :])
                set_input_weight('output', value[3 * hs:, :])
                del state_dict[name]
            elif prefix + 'weight_hh' in name:
                set_hidden_weight('input', value[:hs, :])
                set_hidden_weight('forget', value[hs:hs * 2, :])
                set_hidden_weight('cell', value[2 * hs:hs * 3, :])
                set_hidden_weight('output', value[3 * hs:, :])
                del state_dict[name]

        super(_QuantLSTMLayer, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


class QuantRecurrentStackBase(nn.Module):

    def __init__(
            self,
            layer_impl,
            input_size: int,
            hidden_size: int,
            num_layers: int,
            bidirectional: bool,
            io_quant,
            shared_input_hidden_weights: bool,
            return_quant_tensor: bool,
            **kwargs):
        super(QuantRecurrentStackBase, self).__init__()
        if shared_input_hidden_weights and not bidirectional:
            raise RuntimeError("Shared input-hidden weights requires bidirectional=True.")
        if return_quant_tensor and io_quant is None:
            raise RuntimeError("return_quant_tensor=True requires io_quant != None.")

        self.num_directions = 2 if bidirectional else 1
        layers = []
        # Add io_quant to kwargs. This allows easy overwriting during sharing
        kwargs['io_quant'] = io_quant
        for layer in range(num_layers):
            layer_input_size = input_size if layer == 0 else hidden_size * self.num_directions
            quantize_output_only = bool(layer)
            # return_quant_tensor is required for bias quantization of internal layers
            layer_return_quant_tensor = return_quant_tensor or layer < num_layers - 1
            directions = []
            left_to_right = layer_impl(
                input_size=layer_input_size,
                hidden_size=hidden_size,
                reverse_input=False,
                quantize_output_only=quantize_output_only,
                shared_input_hidden_weights=shared_input_hidden_weights,
                return_quant_tensor=layer_return_quant_tensor,
                **kwargs)
            directions.append(left_to_right)
            # Update kwargs with shared quantizers. Overwrite io_quant and any
            # other quantizer that should be shared. The quantizers of the
            # first left-to-right layer are shared to all directions and layers
            kwargs.update(**left_to_right.quantizers_to_share)
            if bidirectional:
                shared_weights = left_to_right.weights_to_share
                right_to_left = layer_impl(
                    input_size=layer_input_size,
                    hidden_size=hidden_size,
                    reverse_input=True,
                    quantize_output_only=quantize_output_only,
                    shared_input_hidden_weights=shared_input_hidden_weights,
                    return_quant_tensor=layer_return_quant_tensor,
                    **shared_weights,
                    **kwargs)
                directions.append(right_to_left)
            layers.append(nn.ModuleList(directions))
        self.layers = nn.ModuleList(layers)

    def forward(self, inp, hx=None):
        output_states = []
        for l, layer in enumerate(self.layers):
            dir_outputs, dir_states = [], []
            for d, direction in enumerate(layer):
                layer_state = hx[2 * l + d] if hx is not None else hx
                out, out_state = direction(inp, layer_state)
                dir_outputs += [out]
                dir_states += [out_state]
            if len(dir_outputs) > 1:
                out = torch.cat(dir_outputs, dim=-1)
                output_states += [torch.cat(dir_states, dim=0)]
            else:
                out = dir_outputs[0]
                output_states += [dir_states[0]]
            inp = out
        if len(output_states) > 1:
            output_states = torch.cat(output_states, dim=0)
        else:
            output_states = output_states[0]
        return out, output_states

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        for name, value in list(state_dict.items()):
            for index in range(len(self.layers)):
                layer_name = f'_l{index}'
                reverse_layer_name = layer_name + '_reverse'
                if reverse_layer_name in name:
                    param_name = name[len(prefix):-len(reverse_layer_name)]
                    state_dict[f'{prefix}layers.{index}.1.{param_name}'] = value
                    del state_dict[name]
                elif layer_name in name:
                    param_name = name[len(prefix):-len(layer_name)]
                    state_dict[f'{prefix}layers.{index}.0.{param_name}'] = value
                    del state_dict[name]
        super(QuantRecurrentStackBase, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


class QuantRNN(QuantRecurrentStackBase):

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            num_layers: int = 1,
            nonlinearity: str = 'tanh',
            bias: bool = True,
            batch_first: bool = False,
            bidirectional: bool = False,
            weight_quant=Int8WeightPerTensorFloat,
            bias_quant=Int32Bias,
            io_quant=Int8ActPerTensorFloat,
            gate_acc_quant=Int8ActPerTensorFloat,
            shared_input_hidden_weights=False,
            return_quant_tensor: bool = False,
            **kwargs):
        super(QuantRNN, self).__init__(
            layer_impl=_QuantRNNLayer,
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            nonlinearity=nonlinearity,
            bias=bias,
            batch_first=batch_first,
            bidirectional=bidirectional,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            io_quant=io_quant,
            gate_acc_quant=gate_acc_quant,
            shared_input_hidden_weights=shared_input_hidden_weights,
            return_quant_tensor=return_quant_tensor,
            **kwargs)


class QuantLSTM(QuantRecurrentStackBase):

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            num_layers: int = 1,
            bias: bool = True,
            batch_first: bool = False,
            bidirectional: bool = False,
            weight_quant=Int8WeightPerTensorFloat,
            bias_quant=Int32Bias,
            io_quant=Int8ActPerTensorFloat,
            gate_acc_quant=Int8ActPerTensorFloat,
            sigmoid_quant=Uint8ActPerTensorFloat,
            tanh_quant=Int8ActPerTensorFloat,
            cell_state_quant=Int8ActPerTensorFloat,
            coupled_input_forget_gates: bool = False,
            cat_output_cell_states=True,
            shared_input_hidden_weights=False,
            shared_intra_layer_weight_quant=False,
            shared_intra_layer_gate_acc_quant=False,
            shared_cell_state_quant=True,
            return_quant_tensor: bool = False,
            **kwargs):
        super(QuantLSTM, self).__init__(
            layer_impl=_QuantLSTMLayer,
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=batch_first,
            bidirectional=bidirectional,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            io_quant=io_quant,
            gate_acc_quant=gate_acc_quant,
            sigmoid_quant=sigmoid_quant,
            tanh_quant=tanh_quant,
            cell_state_quant=cell_state_quant,
            cifg=coupled_input_forget_gates,
            shared_input_hidden_weights=shared_input_hidden_weights,
            shared_intra_layer_weight_quant=shared_intra_layer_weight_quant,
            shared_intra_layer_gate_acc_quant=shared_intra_layer_gate_acc_quant,
            shared_cell_state_quant=shared_cell_state_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        if cat_output_cell_states and cell_state_quant is not None and not shared_cell_state_quant:
            raise RuntimeError("Concatenating cell states requires shared cell quantizers.")
        self.cat_output_cell_states = cat_output_cell_states

    def forward(self, inp, hx=None, cx=None):
        output_hidden_states, output_cell_states = [], []
        for l, layer in enumerate(self.layers):
            dir_outputs, dir_hidden_states, dir_cell_states = [], [], []
            for d, direction in enumerate(layer):
                layer_hidden_state = hx[2 * l + d] if hx is not None else hx
                layer_cell_state = cx[2 * l + d] if cx is not None else cx
                out, out_hidden_state, out_cell_state = direction(inp, layer_hidden_state, layer_cell_state)
                dir_outputs += [out]
                dir_hidden_states += [out_hidden_state]
                dir_cell_states += [out_cell_state]
            if len(dir_outputs) > 1:
                out = torch.cat(dir_outputs, dim=-1)
                output_hidden_states += [torch.cat(dir_hidden_states, dim=0)]
                if self.cat_output_cell_states:
                    output_cell_states += [torch.cat(dir_cell_states, dim=0)]
                else:
                    output_cell_states.extend(dir_cell_states)
            else:
                out = dir_outputs[0]
                output_hidden_states += [dir_hidden_states[0]]
                if self.cat_output_cell_states:
                    output_cell_states += [dir_cell_states[0]]
                else:
                    output_cell_states.extend(dir_cell_states)
            inp = out
        if len(output_hidden_states) > 1:
            output_hidden_states = torch.cat(output_hidden_states, dim=0)
            if self.cat_output_cell_states:
                output_cell_states = torch.cat(output_cell_states, dim=0)
        else:
            output_hidden_states = output_hidden_states[0]
            if self.cat_output_cell_states:
                output_cell_states = output_cell_states[0]
        return out, (output_hidden_states, output_cell_states)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .acc import AccQuantType
from .act import ActQuantType
from .act import QuantInputMixin
from .act import QuantNonLinearActMixin
from .act import QuantOutputMixin
from .base import QuantLayerMixin
from .parameter import BiasQuantType
from .parameter import QuantBiasMixin
from .parameter import QuantWeightMixin
from .parameter import WeightQuantType



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
import os
import random
import warnings

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision

from brevitas.export import export_onnx_qcdq
from brevitas.export import export_torch_qcdq
from brevitas.graph.quantize import preprocess_for_quantize
from brevitas.graph.target.flexml import preprocess_for_flexml_quantize
from brevitas_examples.imagenet_classification.ptq.ptq_common import calibrate
from brevitas_examples.imagenet_classification.ptq.ptq_common import quantize_model
from brevitas_examples.imagenet_classification.ptq.utils import add_bool_arg
from brevitas_examples.imagenet_classification.ptq.utils import get_model_config
from brevitas_examples.imagenet_classification.ptq.utils import get_torchvision_model
from brevitas_examples.imagenet_classification.utils import generate_dataloader
from brevitas_examples.imagenet_classification.utils import SEED
from brevitas_examples.imagenet_classification.utils import validate

# Ignore warnings about __torch_function__
warnings.filterwarnings("ignore")

model_names = sorted(
    name for name in torchvision.models.__dict__ if name.islower() and not name.startswith("__") and
    callable(torchvision.models.__dict__[name]) and not name.startswith("get_"))

parser = argparse.ArgumentParser(description='PyTorch ImageNet PTQ Validation')
parser.add_argument(
    '--calibration-dir',
    required=True,
    help='Path to folder containing Imagenet calibration folder')
parser.add_argument(
    '--validation-dir', required=True, help='Path to folder containing Imagenet validation folder')
parser.add_argument(
    '--workers', default=8, type=int, help='Number of data loading workers (default: 8)')
parser.add_argument(
    '--batch-size-calibration',
    default=64,
    type=int,
    help='Minibatch size for calibration (default: 64)')
parser.add_argument(
    '--batch-size-validation',
    default=256,
    type=int,
    help='Minibatch size for validation (default: 256)')
parser.add_argument(
    '--export-dir', default='.', type=str, help='Directory where to store the exported models')
parser.add_argument('--gpu', default=None, type=int, help='GPU id to use (default: None)')
parser.add_argument(
    '--calibration-samples', default=1000, type=int, help='Calibration size (default: 1000)')
parser.add_argument(
    '--model-name',
    default='resnet18',
    metavar='ARCH',
    choices=model_names,
    help='model architecture: ' + ' | '.join(model_names) + ' (default: resnet18)')
parser.add_argument(
    '--target-backend',
    default='generic',
    choices=['generic', 'layerwise', 'flexml'],
    help='Backend to target for quantization (default: generic)')
parser.add_argument(
    '--scale-factor-type',
    default='float32',
    choices=['float32', 'po2'],
    help='Type for scale factors (default: float32)')
parser.add_argument(
    '--act-bit-width', default=8, type=int, help='Activations bit width (default: 8)')
parser.add_argument(
    '--weight-bit-width', default=8, type=int, help='Weights bit width (default: 8)')

parser.add_argument(
    '--bias-bit-width',
    default='int32',
    choices=['int32', 'int16'],
    help='Bias bit width (default: int32)')
parser.add_argument(
    '--act-quant-type',
    default='symmetric',
    choices=['symmetric', 'asymmetric'],
    help='Activation quantization type (default: symmetric)')
parser.add_argument(
    '--graph-eq-iterations',
    default=20,
    type=int,
    help='Numbers of iterations for graph equalization (default: 20)')
parser.add_argument(
    '--act-quant-percentile',
    default=99.999,
    type=float,
    help='Percentile to use for stats of activation quantization (default: 99.999)')
parser.add_argument(
    '--export-onnx-qcdq', action='store_true', help='If true, export the model in onnx qcdq format')
parser.add_argument(
    '--export-torch-qcdq',
    action='store_true',
    help='If true, export the model in torch qcdq format')
add_bool_arg(
    parser,
    'scaling-per-output-channel',
    default=True,
    help='Weight scaling per output channel (default: enabled)')
add_bool_arg(
    parser, 'bias-corr', default=True, help='Bias correction after calibration (default: enabled)')
add_bool_arg(
    parser,
    'graph-eq-merge-bias',
    default=True,
    help='Merge bias when performing graph equalization (default: enabled)')
add_bool_arg(
    parser,
    'weight-narrow-range',
    default=True,
    help='Narrow range for weight quantization (default: enabled)')


def main():
    args = parser.parse_args()

    random.seed(SEED)
    np.random.seed(SEED)
    torch.manual_seed(SEED)

    config = (
        f"{args.model_name}_"
        f"{args.target_backend}_"
        f"{args.scale_factor_type}_"
        f"a{args.act_bit_width}"
        f"w{args.weight_bit_width}_"
        f"{'weight_narrow_range_' if args.weight_narrow_range else ''}"
        f"{args.bias_bit_width}bias_"
        f"{'per_channel' if args.scaling_per_output_channel else 'per_tensor'}_"
        f"{args.act_quant_type}_"
        f"{'bc_' if args.bias_corr else ''}"
        f"{args.graph_eq_iterations}geiters_"
        f"{'mb_' if args.graph_eq_merge_bias else ''}"
        f"{args.act_quant_percentile}percentile")

    print(
        f"Model: {args.model_name} - "
        f"Target backend: {args.target_backend} - "
        f"Quantization type: {args.scale_factor_type} - "
        f"Activation bit width: {args.act_bit_width} - "
        f"Weight bit width: {args.weight_bit_width} - "
        f"Weight narrow range: {args.weight_narrow_range} - "
        f"Bias bit width: {args.bias_bit_width} - "
        f"Per-channel scale factors: {args.scaling_per_output_channel} - "
        f"Activation quant type: {args.act_quant_type} - "
        f"Bias Correction Enabled: {args.bias_corr} - "
        f"Iterations for graph equalization: {args.graph_eq_iterations} - "
        f"Merge bias in graph equalization: {args.graph_eq_merge_bias} - "
        f"Activation Quant Percentile: {args.act_quant_percentile}")

    # Get model-specific configurations about input shapes and normalization
    model_config = get_model_config(args.model_name)

    # Generate calibration and validation dataloaders
    resize_shape = model_config['resize_shape']
    center_crop_shape = model_config['center_crop_shape']
    inception_preprocessing = model_config['inception_preprocessing']
    calib_loader = generate_dataloader(
        args.calibration_dir,
        args.batch_size_calibration,
        args.workers,
        resize_shape,
        center_crop_shape,
        args.calibration_samples,
        inception_preprocessing)
    val_loader = generate_dataloader(
        args.validation_dir,
        args.batch_size_validation,
        args.workers,
        resize_shape,
        center_crop_shape,
        inception_preprocessing=inception_preprocessing)

    # Get the model from torchvision
    model = get_torchvision_model(args.model_name)

    # Preprocess the model for quantization
    if args.target_backend == 'flexml':
        # flexml requires static shapes, pass a representative input in
        img_shape = model_config['center_crop_shape']
        model = preprocess_for_flexml_quantize(
            model,
            torch.ones(1, 3, img_shape, img_shape),
            equalize_iters=args.graph_eq_iterations,
            equalize_merge_bias=args.graph_eq_merge_bias)
    elif args.target_backend == 'generic' or args.target_backend == 'layerwise':
        model = preprocess_for_quantize(
            model,
            equalize_iters=args.graph_eq_iterations,
            equalize_merge_bias=args.graph_eq_merge_bias)
    else:
        raise RuntimeError(f"{args.target_backend} backend not supported.")

    # Define the quantized model
    quant_model = quantize_model(
        model,
        backend=args.target_backend,
        act_bit_width=args.act_bit_width,
        weight_bit_width=args.weight_bit_width,
        weight_narrow_range=args.weight_narrow_range,
        bias_bit_width=args.bias_bit_width,
        scaling_per_output_channel=args.scaling_per_output_channel,
        act_quant_percentile=args.act_quant_percentile,
        act_quant_type=args.act_quant_type,
        scale_factor_type=args.scale_factor_type)

    # If available, use the selected GPU
    if args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        quant_model = quant_model.cuda(args.gpu)
        cudnn.benchmark = False

    # Calibrate the quant_model on the calibration dataloader
    print("Starting calibration")
    calibrate(calib_loader, quant_model, args.bias_corr)

    # Validate the quant_model on the validation dataloader
    print("Starting validation")
    validate(val_loader, quant_model)

    if args.export_onnx_qcdq or args.export_torch_qcdq:
        # Generate reference input tensor to drive the export process
        model_config = get_model_config(args.model_name)
        center_crop_shape = model_config['center_crop_shape']
        img_shape = center_crop_shape
        device, dtype = next(model.parameters()).device, next(model.parameters()).dtype
        ref_input = torch.ones(1, 3, img_shape, img_shape, device=device, dtype=dtype)

        export_name = os.path.join(args.export_dir, config)
        if args.export_onnx_qcdq:
            export_name = export_name + '.onnx'
            export_onnx_qcdq(model, ref_input, export_name)
        if args.export_torch_qcdq:
            export_name = export_name + '.pt'
            export_torch_qcdq(model, ref_input, export_name)


if __name__ == '__main__':
    main()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import Optional, Type, Union
from warnings import warn

from brevitas.inject import ExtendedInjector
from brevitas.inject import Injector
from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjector
from brevitas.proxy.parameter_quant import BiasQuantProxyProtocol
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector
from brevitas.proxy.parameter_quant import WeightQuantProxyProtocol
from brevitas.quant import NoneBiasQuant
from brevitas.quant import NoneWeightQuant
from brevitas.quant_tensor import QuantTensor

from .base import QuantProxyMixin

WeightQuantType = Union[WeightQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]
BiasQuantType = Union[BiasQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]


class QuantWeightMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, weight_quant: Optional[WeightQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=weight_quant,
            proxy_protocol=WeightQuantProxyProtocol,
            none_quant_injector=NoneWeightQuant,
            kwargs_prefix='weight_',
            proxy_prefix='weight_',
            **kwargs)

    @property
    @abstractmethod
    def output_channel_dim(self) -> int:
        pass

    @property
    def is_weight_quant_enabled(self):
        return self.weight_quant.is_quant_enabled

    @property
    def is_quant_weight_narrow_range(self):
        return self.weight_quant.is_narrow_range

    @property
    def is_quant_weight_signed(self):
        return self.weight_quant.is_signed

    @property
    def weight_quant_requires_quant_input(self):
        return self.weight_quant.requires_quant_input

    def quant_weight(self, quant_input: Optional[QuantTensor] = None):
        if self.weight_quant_requires_quant_input:
            if quant_input is None:
                input_bit_width = self.quant_input_bit_width()
                input_is_signed = self.is_quant_input_signed
            else:
                input_bit_width = quant_input.bit_width
                input_is_signed = quant_input.signed
            assert input_bit_width is not None, "Input bit-width needs to be specified."
            assert input_is_signed is not None, "Input sign needs to be specified."
            return self.weight_quant(self.weight, input_bit_width, input_is_signed)
        return self.weight_quant(self.weight)

    def int_weight(self, float_datatype=False):
        return self.quant_weight().int(float_datatype)

    def quant_weight_scale(self):
        scale = self.quant_weight().scale
        return scale

    def quant_weight_zero_point(self):
        scale = self.quant_weight().zero_point
        return scale

    def quant_weight_bit_width(self):
        bit_width = self.quant_weight().bit_width
        return bit_width

    def register_parameter(self, name, value):
        super(QuantWeightMixin, self).register_parameter(name, value)
        if hasattr(self, 'weight_quant') and name == 'weight':
            self.weight_quant.init_tensor_quant()
            self.weight_quant.to(self.weight.device)


class QuantBiasMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            bias_quant: Optional[BiasQuantType],
            cache_inference_bias: bool = False,
            **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=bias_quant,
            proxy_protocol=BiasQuantProxyProtocol,
            none_quant_injector=NoneBiasQuant,
            kwargs_prefix='bias_',
            proxy_prefix='bias_',
            **kwargs)
        self.cache_inference_quant_bias = cache_inference_bias
        self._cached_bias = None

    @property
    def is_bias_quant_enabled(self):
        return self.bias_quant.is_quant_enabled

    @property
    def is_quant_bias_narrow_range(self):
        if self.bias is None:
            return None
        return self.bias_quant.is_narrow_range

    @property
    def is_quant_bias_signed(self):
        if self.bias is None or not self.is_bias_quant_enabled:
            return None
        return self.bias_quant.is_signed

    def int_bias(self, float_datatype=False):
        if self.bias is None or not self.is_bias_quant_enabled:
            return None
        quant_bias = self.quant_bias()
        return quant_bias.int(float_datatype=float_datatype)

    def quant_bias(self):
        if self.bias is None:
            return None
        scale = self.quant_bias_scale()
        bit_width = self.quant_bias_bit_width()
        quant_bias = self.bias_quant(self.bias, scale, bit_width)
        return quant_bias

    def quant_bias_scale(self):
        if self.bias is None or not self.is_bias_quant_enabled:
            return None
        if not self.bias_quant.requires_input_scale and not self.bias_quant.requires_input_bit_width:
            return self.bias_quant(self.bias).scale
        else:
            if self._cached_bias is None:
                raise RuntimeError(
                    "No quant bias cache found, set cache_inference_quant_bias=True and run an "
                    "inference pass first")
            if self.training:
                warn("Cached quant bias scale is being used in training mode.")
            return self._cached_bias.scale

    def quant_bias_zero_point(self):
        if self.bias is None:
            return None
        if not self.bias_quant.requires_input_scale and not self.bias_quant.requires_input_bit_width:
            return self.bias_quant(self.bias).zero_point
        else:
            if self._cached_bias is None:
                raise RuntimeError(
                    "No quant bias cache found, set cache_inference_quant_bias=True and run an "
                    "inference pass first")
            if self.training:
                warn("Cached quant bias zero-point is being used in training mode.")
            return self._cached_bias.bit_width

    def quant_bias_bit_width(self):
        if self.bias is None or not self.is_bias_quant_enabled:
            return None
        if not self.bias_quant.requires_input_scale and not self.bias_quant.requires_input_bit_width:
            return self.bias_quant(self.bias).bit_width
        else:
            if self._cached_bias is None:
                raise RuntimeError(
                    "No quant bias cache found, set cache_inference_quant_bias=True and run an "
                    "inference pass first")
            if self.training:
                warn("Cached quant bias bit-width is being used in training mode.")
            return self._cached_bias.bit_width

    def register_parameter(self, name, value):
        super(QuantBiasMixin, self).register_parameter(name, value)
        if hasattr(self, 'bias_quant') and name == 'bias':
            self.bias_quant.init_tensor_quant()
            self.bias_quant.to(self.bias.device)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

"""
Implementation of various functions to compute shapes that induce flattening along certain
dimensions of a tensor.
"""

from typing import Tuple

import torch
from torch import Tensor

import brevitas

__all__ = [
    'over_tensor',
    'over_output_channels',
    'over_batch_over_tensor',
    'over_batch_over_output_channels']


@brevitas.jit.script
def over_tensor(x: Tensor) -> int:
    """
    Computes the shape s such that x.view(s) is a flat tensor.

    Args:
        x (Tensor): Input tensor.

    Returns:
        The number -1 corresponding to a flat shape.

    Examples:
        >>> over_tensor(torch.randn([2, 3, 4, 3]))
        -1
    """
    return -1


@brevitas.jit.script
def over_output_channels(x: Tensor) -> Tuple[int, int]:
    """
    Computes the shape s such that x.view(s) is a 2-dim tensor with output channels
    at dimension 0 and any other feature at dimension 1.

    Args:
    x (Tensor): Input tensor with output channels at dimension 0.

    Returns:
        A tuple containing the 2-dim shape.

    Examples:
        >>> over_output_channels(torch.randn([2, 3, 4, 3]))
        (2, -1)
    """
    return x.shape[0], -1


@brevitas.jit.script
def over_batch_over_tensor(x: Tensor) -> Tuple[int, int]:
    """
    Computes the shape s such that x.view(s) is a 2-dim tensor with batches
    at dimension 0 and any other feature at dimension 1.

    Args:
        x (Tensor): Input tensor with batches at dimension 0.

    Returns:
        A tuple containing the 2-dim shape.

    Examples:
        >>> over_batch_over_tensor(torch.randn([2, 3, 4, 3]))
        (2, -1)
    """
    return x.shape[0], -1


@brevitas.jit.script
def over_batch_over_output_channels(x: Tensor):
    """
    Returns a shape s such that x.view(s) is a 3-dim tensor with batches
    at dimension 0, output channels at dimension 1, and any other feature at dimension 2.

    Args:
        x (Tensor): Input tensor with batches at dimension 0 and output channels at dimension 1.

    Returns:
        A tuple containing the 3-dim shape.

    Examples:
        >>> over_batch_over_output_channels(torch.randn([2, 3, 4, 3]))
        (2, 3, -1)
    """
    return x.shape[0], x.shape[1], -1
